from unittest import TestCase

from pandas_ta_quant import pd
from pandas_ta_quant.technical_analysis import ta_ewma_covariance, ta_mgarch_covariance
from pandas_ta_quant_test.config import DF_TEST_MULTI_ROW, DF_TEST_MULTI


class TestCovariances(TestCase):

    def test_ewma_covariance(self):
        df = DF_TEST_MULTI._["Close"]
        cov = ta_ewma_covariance(df)
        # print(cov.tail())

    def test_ewma_covariance_multiindex_row(self):
        df = pd.concat([DF_TEST_MULTI.add_multi_index("A", axis=0), DF_TEST_MULTI.add_multi_index("B", axis=0)], axis=0)
        cov = ta_ewma_covariance(df)
        pd.testing.assert_frame_equal(cov.loc["A"], cov.loc["B"])

    def test_mgarch_covariance(self):
        df = DF_TEST_MULTI[-40:]._["Close"]
        cov = ta_mgarch_covariance(df)
        # print(cov)

        # ewma cov result:
        #.00002107488e-05 -1.341822e-05
        #  (Close, gld) -1.341822e-05  2.034593e-05
        self.assertGreater(cov.iloc[-2, -2], 2.107488e-05)
        self.assertLess(cov.iloc[-1, -2], 0)



from pandas_ml_common import pd as _pd, suitable_kwargs as _select_kwargs
from pandas_ta_quant._decorators import *
from .multi_object import *
from .single_object import *
from .time import *
from ..encoders.candles import ta_candle_category, ta_candles_as_culb


# this is executing all technical indicators into one huge data frame
@for_each_top_level_row
@for_each_top_level_column
def ta_all(df: _pd.DataFrame,
           close_only: bool = False,
           open: str = "Open",
           high: str = "High",
           low: str = "Low",
           close: str = "Close",
           volume: str = "Volume",
           open_interest: str = "Open Interest"
           ):
    assert isinstance(df, _pd.DataFrame), "Data needs to be a frame not a Series"
    df_res = None

    # loop all indicator functions
    for indicator_function in dir(single_object):
        if indicator_function.startswith("ta_"):
            func = getattr(single_object, indicator_function)

            if hasattr(func, '_is_timeconsuming'):
                continue

            if df_res is None:
                df_res = func(df[close]).to_frame().add_multi_index(indicator_function[3:])
            else:
                df_res = df_res.inner_join(func(df[close]), prefix=indicator_function[3:], force_multi_index=True)

    for indicator_function in dir(time):
        if indicator_function.startswith("ta_"):
            func = getattr(time, indicator_function)

            if hasattr(func, '_is_timeconsuming'):
                continue

            df_res = df_res.inner_join(func(df), prefix=indicator_function[3:], force_multi_index=True)

    if close_only:
        return df_res

    for indicator_function in dir(multi_object):
        if indicator_function.startswith("ta_"):
            func = getattr(multi_object, indicator_function)

            if hasattr(func, '_is_timeconsuming'):
                continue

            _df = func(df, **_select_kwargs(func, open=open, high=high, low=low, close=close, volume=volume,
                                            open_interest=open_interest))
            df_res = df_res.inner_join(_df, prefix=indicator_function[3:], force_multi_index=True)

    # add corner cases i.e. candle stock encodings as indicators
    df_res = df_res.inner_join(ta_candle_category(df, open=open, high=high, low=low, close=close), prefix='candle_category', force_multi_index=True)
    df_res = df_res.inner_join(ta_candles_as_culb(df, open=open, high=high, low=low, close=close, volume=volume), prefix='culb', force_multi_index=True)

    return df_res


"""
this module basically re-implements all oscillators from TA-Lib:
  https://mrjbq7.github.io/ta-lib/func_groups/momentum_indicators.html
"""

"""
TODO add this missing indicators

AROONOSC - Aroon Oscillator
real = AROONOSC(high, low, timeperiod=14)
Learn more about the Aroon Oscillator at tadoc.org.

CMO - Chande Momentum Oscillator
real = CMO(close, timeperiod=14)
Learn more about the Chande Momentum Oscillator at tadoc.org.

MFI - Money Flow Index
real = MFI(high, low, close, volume, timeperiod=14)
Learn more about the Money Flow Index at tadoc.org.

STOCH - Stochastic
slowk, slowd = STOCH(high, low, close, fastk_period=5, slowk_period=3, slowk_matype=0, slowd_period=3, slowd_matype=0)
Learn more about the Stochastic at tadoc.org.

"""


import pandas as _pd

# create convenient type hint
from pandas_ml_common import Typing as _t
from pandas_ta_quant._decorators import *


@for_each_top_level_row
def ta_volume_as_time(df: _t.PatchedPandas, volume="Volume"):
    if df.ndim > 1:
        res = df.copy()
        res.index = df[volume].cumsum()
        return res
    else:
        return _pd.Series(df.index, index=df.cumsum())




from unittest import TestCase

import numpy as np
import pandas as pd

from pandas_ta_quant._utils import index_of_bucket, rolling_apply


class TestUtils(TestCase):

    def test_index_of_bucket(self):

        a = index_of_bucket(0, np.array([1, 2, 3]))
        b = index_of_bucket(1.2, np.array([1, 2, 3]))
        c = index_of_bucket(2.1, np.array([1, 2, 3]))
        d = index_of_bucket(3, np.array([1, 2, 3]))

        self.assertEqual(0, a)
        self.assertEqual(1, b)
        self.assertEqual(2, c)
        self.assertEqual(3, d)

    def test_roll_apply(self):
        s = pd.Series([1,2,3,4,5])

        m1 = rolling_apply(s, 3, lambda x: x.mean(), "foo")
        np.testing.assert_array_almost_equal(m1.values.reshape(len(s)), [np.nan, np.nan, 2, 3, 4])

        m4 = rolling_apply(s, 4, lambda x: x.mean(), "foo")
        np.testing.assert_array_almost_equal(m4.values.reshape(len(s)), [np.nan, np.nan, np.nan, 2.5, 3.5])

        m5 = rolling_apply(s, 5, lambda x: x.mean(), "foo")
        np.testing.assert_array_almost_equal(m5.values.reshape(len(s)), [np.nan, np.nan, np.nan, np.nan, 3])

        m2 = rolling_apply(s, 3, lambda x: x.mean(), "bar", center=True)
        np.testing.assert_array_almost_equal(m2.values.reshape(len(s)), [np.nan, 2, 3, 4, np.nan])

        m3 = rolling_apply(s, 5, lambda x: x.mean(), "baz", center=True)
        np.testing.assert_array_almost_equal(m3.values.reshape(len(s)), [np.nan, np.nan, 3, np.nan, np.nan])


from unittest import TestCase

import numpy as np

from pandas_ta_quant.technical_analysis import ta_ma_ratio, ta_returns, ta_cumret, ta_log_returns, ta_performance, \
    ta_cumlogret, ta_realative_candles
from pandas_ta_quant_test.config import DF_TEST


class TestNormalizer(TestCase):

    def test_relative_candles(self):
        df = DF_TEST[-2:].copy()
        relative = ta_realative_candles(df, volume=None)

        np.testing.assert_array_almost_equal(np.array([0.002469,  0.002021,  0.002533, -0.002829]),
                                             relative.values[-1])

    def test_ma_ratio(self):
        df = DF_TEST

        ma_ration_scaled = ta_ma_ratio(df)
        print(ma_ration_scaled.columns)
        self.assertTrue(True)

    def test_cumret(self):
        c = DF_TEST["Close"]
        perf = (1 + c.pct_change()).cumprod().rename("perf")
        cumrets = ta_cumret(ta_returns(c, 2), 2).rename("cumrets")

        results = perf.to_frame().join(cumrets).dropna()
        np.testing.assert_array_almost_equal(results["cumrets"].values[-20:-2], results["perf"].values[-20:-2])

    def test_cumlogret(self):
        c = DF_TEST["Close"]
        perf = ta_performance(c)
        logret = ta_log_returns(c)
        cumrets = ta_cumlogret(logret)

        np.testing.assert_array_almost_equal(cumrets.values, perf.values)


from unittest import TestCase

from pandas_ta_quant import pd
from pandas_ta_quant_test.config import DF_TEST


class TestMultiIndexRowDf(TestCase):

    def test_indicator(self):
        df1 = DF_TEST.copy()
        df2 = DF_TEST.copy()
        df1.index = pd.MultiIndex.from_product([["A"], df1.index.tolist()])
        df2.index = pd.MultiIndex.from_product([["B"], df2.index.tolist()])
        df = pd.concat([df1, df2], axis=0)

        self.assertEqual(df.loc["A"].shape, df.loc["B"].shape)

        res = df["Close"].ta.macd()
        pd.testing.assert_frame_equal(res.loc["A"], res.loc["B"])



from typing import Union as _Union

# create convenient type hint
import pandas as _pd

from pandas_ta_quant._decorators import *

_PANDAS = _Union[_pd.DataFrame, _pd.Series]


@for_each_top_level_row
def ta_inverse(df: _PANDAS) -> _PANDAS:
    return df.apply(lambda col: col * -1 + col.min() + col.max(), raw=True)




from typing import Tuple

import numpy as np
from sortedcontainers import SortedKeyList
from pandas_ta_quant._decorators import *
from pandas_ml_common import Typing


@for_each_top_level_row
def ta_fibbonaci_retracement(df: Typing.PatchedPandas, period=200, patience=3):
    current_min_max = [0, 0]
    most_recent_min_max = [0, 0]
    count = [0, 0]

    def call_after_reset(func):
        nonlocal most_recent_min_max
        nonlocal current_min_max
        nonlocal count

        current_min_max = [0, 0]
        most_recent_min_max = [0, 0]
        count = [0, 0]

        return func()

    def fibonacci(col, fact):
        nonlocal most_recent_min_max
        nonlocal current_min_max
        nonlocal count
        min_max = np.min(col), np.max(col)

        # as long min/max is changing because the window moves we use the currently valid min and max
        # but as soon as min and max stays stable we set this as the new currently valid min and max values
        for i in range(2):
            if min_max[i] == most_recent_min_max[i]:
                if count[i] > patience:
                    current_min_max[i] = min_max[i]
                    count[i] = 0

                count[i] += 1

        min_max_range = current_min_max[1] - current_min_max[0]
        most_recent_min_max = min_max

        valid_min_max_range = min_max_range > 0.001 and current_min_max[0] > 0.001 and current_min_max[1] > 0.001
        return (min_max_range * fact + current_min_max[0]) if valid_min_max_range else np.nan

    retracements = {"fourty":  0.382, "fitfy": 0.5, "sixty": 0.618}
    return pd.DataFrame(
        {k: call_after_reset(lambda: df.rolling(period).apply(lambda col: fibonacci(col, v), raw=True)) for k, v in
         retracements.items()},
        index=df.index
    )


@for_each_top_level_row
def ta_trend_lines(df: Typing.PatchedSeries,
                   edge_periods=3,
                   rescale_digits=4,
                   degrees=(-90, 90),
                   angles=30,
                   rho_digits=2,
                   edge_detect='mean',
                   **kwargs
                   ) -> Tuple[pd.DataFrame, pd.DataFrame]:
    from pandas_ta_quant.technical_analysis.edge_detect import EDGE_DETECTOR
    assert df.ndim == 1 or len(df.columns) == 1, "Trend lines can only be calculated on a series"

    # edge detection
    rescaled = df.ta.rescale((0, 1), digits=rescale_digits)
    edge_or_not = EDGE_DETECTOR[edge_detect](rescaled, period=edge_periods, **kwargs)

    # set up spaces
    x = np.linspace(0, 1, len(rescaled))
    y = rescaled.values.reshape(x.shape)
    edge_x_index = np.arange(0, len(rescaled))[edge_or_not != 0]
    edge_x = x[edge_or_not != 0]
    edge_y = y[edge_or_not != 0]
    thetas = np.deg2rad(np.linspace(*degrees, len(edge_x) if angles is None else angles))

    # pre compute angeles, calculate rho's
    cos_theta = np.cos(thetas)
    sin_theta = np.sin(thetas)

    if angles is None:
        # this matrix operation might be more optimized
        rhos = np.outer(cos_theta, edge_x) + np.outer(sin_theta, edge_y)
    else:
        rhos = np.vstack([edge_x[i] * cos_theta + edge_y[i] * sin_theta for i in range(len(edge_x))]).T

    # round rhos and construct a lookup table to map back from theta/rho to x/y
    rhos, time_value_lookup_table = np.around(rhos, rho_digits), {}

    for index, rho in np.ndenumerate(rhos):
        k = (thetas[index[0]], rho)
        time = df.index[edge_x_index[index[1]]]
        value = df.iloc[edge_x_index[index[1]]]

        if k in time_value_lookup_table:
            time_value_lookup_table[k].add((time, value))
        else:
            time_value_lookup_table[k] = SortedKeyList([(time, value)], key=lambda x: x[0])

    # setup the hugh space (plots nice sinusoid's
    hough_space = pd.DataFrame(rhos, index=thetas)

    # filtering
    unique_rhos = np.unique(hough_space)

    def accumulator(row):
        rhos, counts = np.unique(row, return_counts=True)
        s = pd.Series(0, index=unique_rhos)
        s[rhos] = counts
        return s

    # generate a data frame of counts with shape [angels, rhos]
    accumulated = hough_space.apply(accumulator, axis=1)

    # build lookups for filtering
    theta_indices, rho_indices = np.unravel_index(np.argsort(accumulated.values, axis=None), accumulated.shape)
    touches = []
    distances = []
    points = []

    for i in range(len(theta_indices)):
        tp = (thetas[theta_indices[i]], unique_rhos[rho_indices[i]])

        if tp in time_value_lookup_table:
            p = time_value_lookup_table[tp]
            if len(p) > 1:
                touches.append(len(p))
                distances.append(p[-1][0] - p[0][0] if len(p) > 1 else 0)
                points.append(p)

    line_lookup_table = pd.DataFrame(
        {
            "touch": touches,
            "distance": distances,
            "points": points
        },
        index=range(len(points), 0, -1)
    )

    return accumulated, line_lookup_table


@for_each_top_level_row
def ta_ohl_trend_lines(df: Typing.PatchedPandas, close="Close", high=None, low=None):
    # TODO implement this paper: http://www.meacse.org/ijcar/archives/128.pdf
    #   analog ta_trend_lines
    if df.ndim > 1:
        c = df[close]
        h = df[high] if high is not None else None
        l = df[low] if low is not None else None
    else:
        c, h, l = df, None, None




import os
from unittest import TestCase

from pandas_ml_common_test.notebook_runner import run_all_notebooks

PWD = os.path.dirname(os.path.abspath(__file__))


class TestNotebooks(TestCase):

    def test_all_notebooks(self):
        notebooks_path = os.path.join(PWD, '..', 'examples')
        run_all_notebooks(notebooks_path, assert_nofail=True)



from typing import Union as _Union

# create convenient type hint
import numpy as _np
import numpy as np
import pandas as _pd
import pandas as pd
from scipy.stats import zscore

import pandas_ta_quant.technical_analysis.bands as _bands
from pandas_ml_common import Typing
from pandas_ml_common.utils import cumcount
from pandas_ta_quant._decorators import *
from pandas_ta_quant.technical_analysis.filters import ta_ema as _ema, ta_wilders as _wilders, ta_sma as _sma
from pandas_ta_quant._utils import with_column_suffix as _wcs, rolling_apply
from scipy.stats import linregress

_PANDAS = _Union[_pd.DataFrame, _pd.Series]


@for_each_top_level_row
def ta_mean_returns(df: Typing.PatchedPandas, period=20) -> _PANDAS:
    return _wcs(f"mean_return_{period}", df.pct_change().rolling(period).mean())


@for_each_column
@for_each_top_level_row
def ta_macd(df: Typing.PatchedPandas, fast_period=12, slow_period=26, signal_period=9, relative=True) -> _PANDAS:
    fast = _ema(df, fast_period)
    slow = _ema(df, slow_period)

    macd = (fast / slow - 1) if relative else (fast - slow)
    signal = _ema(macd, signal_period)
    hist = macd - signal
    suffix = f'{fast_period},{slow_period},{signal_period}'

    for label, frame in {f"macd_{suffix}": macd, f"signal_{suffix}": signal, f"histogram_{suffix}": hist}.items():
        frame.name = label

    macd = macd.to_frame() if isinstance(macd, _pd.Series) else macd
    return macd.join(signal).join(hist)


@for_each_top_level_row
def ta_mom(df: _PANDAS, period=12, relative=True) -> _PANDAS:
    return _wcs(f"mom_{period}", df.pct_change(period) if relative else df.diff(period))


@for_each_top_level_row
def ta_roc(df: _PANDAS, period=12) -> _PANDAS:
    return _wcs(f"roc_{period}", df.pct_change(period))


@for_each_top_level_row
def ta_stddev(df: _PANDAS, period=12, nbdev=1, ddof=1, downscale=True) -> _PANDAS:
    return _wcs(f"stddev_{period}", (df.rolling(period).std(ddof=ddof) * nbdev) / (100 if downscale else 1))


@for_each_top_level_row
def ta_rsi(df: _PANDAS, period=12):
    returns = df.diff()

    pos = _wilders(returns.clip(lower=0), period)
    neg = _wilders(_np.abs(returns.clip(upper=0)), period)

    return _wcs(f"rsi_{period}", pos / (pos + neg), returns)


@for_each_top_level_row
def ta_apo(df: _PANDAS, fast_period=12, slow_period=26, exponential=False, relative=True) -> _PANDAS:
    fast = _ema(df, fast_period) if exponential else _sma(df, fast_period)
    slow = _ema(df, slow_period) if exponential else _sma(df, slow_period)
    apo = (fast / slow) if relative else (fast - slow)
    return _wcs(f"apo_{fast_period},{slow_period},{int(exponential)}", apo, df)


@for_each_top_level_row
def ta_trix(df: _PANDAS, period=30) -> _PANDAS:
    return _wcs(f"trix_{period}", _ema(_ema(_ema(df, period), period), period).pct_change() * 100, df)


@for_each_top_level_row
def ta_ppo(df: _pd.DataFrame, fast_period=12, slow_period=26, exponential=True) -> _PANDAS:
    fast = _ema(df, period=fast_period) if exponential else _sma(df, period=fast_period)
    slow = _ema(df, period=slow_period) if exponential else _sma(df, period=slow_period)
    return _wcs(f"ppo_{fast_period},{slow_period},{int(exponential)}", (fast - slow) / slow, df)


@for_each_top_level_row
def ta_zscore(df: _PANDAS, period=20, ddof=1, downscale=True):
    res = df.rolling(period).apply(lambda c: zscore(c, ddof=ddof)[-1] / (4 if downscale else 1))

    return _wcs(f"z_{period}", res)


@for_each_top_level_row
@for_each_column
def ta_up_down_volatility_ratio(df: Typing.PatchedSeries, period=60, normalize=True, setof_date=False):
    returns = df.pct_change() if normalize else df
    std = _pd.DataFrame({
        "std+": returns[returns > 0].rolling(period).std(),
        "std-": returns[returns < 0].rolling(period).std()
    }, index=returns.index).fillna(method="ffill")

    ratio = (std["std+"] / std["std-"] - 1).rename("std +/-")

    # eventually we can off set the date such that we can fake one continuous data frame
    if setof_date:
        # +7 -1 binds us approximately to the same week day
        ratio.index = ratio.index - _pd.DateOffset(days=(ratio.index[-1] - ratio.index[0]).days + 7 - 1)

    return ratio


@for_each_top_level_row
@for_each_column
def ta_poly_coeff(df: _PANDAS, period=60, degree=2):
    from pandas_ml_common.utils import ReScaler
    from pandas_ml_common import np_nans

    x = _np.linspace(0, 1, period)
    v = df.values
    res = np_nans((len(df), degree + 1))

    for i in range(period, len(df)):
        y =  v[i - period:i]
        rescaler = ReScaler((y.min(), y.max()), (-1, 1))
        p, err, _, _, _ = _np.polyfit(x, rescaler(y), degree, full=True)
        res[i] = _np.hstack([p[:-1], err / period])

    return _pd.DataFrame(res, index=df.index, columns=[*[f'b{b}' for b in range(0, degree)], "mse"])


@for_each_top_level_row
def ta_sharpe_ratio(df: _PANDAS, period=60, ddof=1, risk_free=0, is_returns=False):
    returns = df if is_returns else df.pct_change()
    mean_returns = returns.rolling(period).mean()
    sigma = returns.rolling(period).std(ddof=ddof)
    return (mean_returns - risk_free) / sigma


@for_each_top_level_row
def ta_sortino_ratio(df: _PANDAS, period=60, ddof=1, risk_free=0, is_returns=False):
    returns = df if is_returns else df.pct_change()
    mean_returns = returns.rolling(period).mean()
    sigma = returns.clip(upper=risk_free).rolling(period).std(ddof=ddof)
    return (mean_returns - risk_free) / sigma


@for_each_top_level_row
@for_each_column
def ta_draw_down(s: _PANDAS, return_dates=False, return_duration=False):
    df = s.copy() if s.ndim > 1 else s.to_frame()

    max_peaks_idx = s.expanding(min_periods=1).apply(lambda x: x.argmax()).fillna(0).astype(int).rename("iloc")
    duration_per_peak_iloc = cumcount(max_peaks_idx.to_frame().reset_index()["iloc"]).rename("duration").to_frame()
    duration_per_peak_iloc["iloc_start"] = duration_per_peak_iloc.index.to_series().shift(1).fillna(0).astype(int)
    duration_per_peak_iloc = duration_per_peak_iloc[duration_per_peak_iloc["duration"] > 1]
    df['max_peaks_idx'] = max_peaks_idx

    nw_peaks = pd.Series(s.iloc[max_peaks_idx.values].values, index=s.index)

    df['dd'] = ((s - nw_peaks) / nw_peaks)
    df['mdd'] = df.groupby('max_peaks_idx').dd.apply(lambda x: x.expanding(min_periods=1).apply(lambda y: y.min())).fillna(0)
    df = df.drop('max_peaks_idx', axis=1)

    if return_dates:
        df = df.join(
            pd.DataFrame(
                {
                    "dd_start" : s.index[duration_per_peak_iloc["iloc_start"]],
                    "dd_end" : s.index[duration_per_peak_iloc.index],
                },
                index=s.index[duration_per_peak_iloc.index]
            )
        )

    if return_duration:
        df = df.join(
            pd.Series(
                duration_per_peak_iloc["duration"].values,
                index=s.index[duration_per_peak_iloc.index],
                name="duration"
            )
        )

    return df.drop(s.columns[1] if s.ndim > 1 else s.name, axis=1)


@for_each_top_level_row
def ta_ma_decompose(df: Typing.PatchedPandas, period=50, boost_ma=10):
    ma = _sma(df, period=period)
    base = ma.pct_change() * boost_ma
    ratio = df / ma - 1

    if isinstance(ratio, Typing.Series):
        ratio.name = f'{df.name}_ma_ratio'

    return base.to_frame().join(ratio)


@for_each_top_level_row
def ta_ma_ratio(df: Typing.PatchedPandas, period=12, exponential='wilder') -> _PANDAS:
    if exponential is True:
        return (1 / _ema(df)) * df.values - 1
    if exponential == 'wilder':
        return (1 / _wilders(df)) * df.values - 1
    else:
        return (1 / _sma(df)) * df.values - 1


@for_each_top_level_row
def ta_std_ret_bands_indicator(df: _pd.DataFrame, period=12, stddevs=[2.0], ddof=1, lag=1, scale_lag=True, include_mean=True, inculde_width=True) -> _PANDAS:
    columns = ["width"]
    res = _bands.ta_std_ret_bands(df, period, stddevs, ddof, lag, scale_lag, inculde_width=inculde_width, include_mean=include_mean)

    if isinstance(res.columns, _pd.MultiIndex):
        columns = [c for c in res.columns.to_list() if c[-1] in columns]

    return res[columns]


@for_each_top_level_row
def ta_bbands_indicator(df: _PANDAS, period=12, stddev=2.0, ddof=1) -> _PANDAS:
    columns = ["width", "z-score", "quantile"]
    res = _bands.ta_bbands(df, period, stddev, ddof)

    if isinstance(res.columns, _pd.MultiIndex):
        columns = [c for c in res.columns.to_list() if c[-1] in columns]

    return res[columns]


@for_each_top_level_row
@for_each_column
def ta_slope(df: _PANDAS, period=12):
    x = _np.arange(period)
    return _wcs(f"slope_{period}", df.rolling(period).apply(lambda y: linregress(x, y).slope))


@for_each_top_level_row
@for_each_column
def ta_potential_turning_point(
        df: _PANDAS,
        period=120,
        point_threshold=2,
        degrees=(-90, 90),
        angles=30,
        realtive=True,
        rho_digits=2,
        rescale_digits=4,
        edge_detector='naive',
        **kwargs):
    from pandas_ta_quant.technical_analysis.edge_detect import EDGE_DETECTOR
    from pandas_ta_quant.technical_analysis.normalizer import ta_rescale
    kwargs = {k.replace("edge_", ""): v for k, v in kwargs.items()}
    edge_or_not = EDGE_DETECTOR[edge_detector](df, **kwargs)

    def calc_edge_count(df_edge_or_not):
        df = df_edge_or_not.iloc[:, 0]
        edge_or_not = df_edge_or_not.iloc[:, 1].values

        # set up spaces
        x = _np.linspace(0, 1, len(df))
        y = ta_rescale(df, (0, 1), digits=rescale_digits)

        # select only edge points pus the last point as we want to do like this last point is an edge point
        mask = edge_or_not != 0
        if mask.sum() <= 3: return 0
        mask[-1] = True

        edge_x, edge_y = x[mask], y[mask]
        thetas = _np.deg2rad(_np.linspace(*degrees, len(edge_x) if angles is None else angles))

        # pre compute angeles, calculate rho's -> 2d Matrix [angles, edge_points]
        cos_theta, sin_theta = _np.cos(thetas), _np.sin(thetas)
        rhos = _np.around(_np.outer(cos_theta, edge_x) + _np.outer(sin_theta, edge_y), rho_digits)

        # get the counts of all unique rhos of the last index and return the count
        rhos_of_interest = _np.unique(rhos[:, -1])
        counts = {roi: (rhos[:, :-1] == roi).sum() for roi in rhos_of_interest}

        return sum([v for v in counts.values() if v > point_threshold]) / ((period * len(thetas)) if realtive else 1)

    return rolling_apply(pd.concat([df, edge_or_not], axis=1), period, calc_edge_count, names=f"PPL_{period}")





from csv import reader
from unittest import TestCase

import numpy
import numpy as np
import pandas as pd

from pandas_ta_quant.portfolio import Portfolio, TargetQuantity, TargetWeight
from pandas_ta_quant_test.config import CSV_TRADES


class TestPortfolio(TestCase):

    def get_trades(self):
        with open(CSV_TRADES) as csv:
            csv_reader = reader(csv)
            header = next(csv_reader)

            for trade in csv_reader:
                if trade[0] == 'VXX':
                    strategy = 'intraday'
                elif trade[0] == 'SPY':
                    strategy = 'slow'
                elif trade[0] == 'WPG':
                    strategy = 'options'
                else:
                    strategy = None

                yield trade[0], trade[1], trade[2], trade[4], trade[5], float(trade[7]), float(trade[8]), None, trade[6], strategy, trade[3], float(trade[-1])

    def test_orders_with_nav(self):
        p = Portfolio()

        for trade in self.get_trades():
            p.trade(*trade)

        pf = p.get_current_portfolio()

        numpy.testing.assert_array_almost_equal(
            pf.loc["WPG"].sum().values,
            np.array([198., 465., -2.973, -np.inf]),
            decimal=3
        )
        numpy.testing.assert_array_almost_equal(
            pf.loc["SPY"].sum().values,
            np.array([0., -20., -0.025, -np.inf]),
            decimal=3
        )
        numpy.testing.assert_array_almost_equal(
            pf.loc["VXX"].sum().values,
            np.array([0., 39., -0.001, -np.inf]),
            decimal=3
        )
        numpy.testing.assert_array_almost_equal(
            pf.loc["USD"].sum().values,
            np.array([-484.0 -2.999, -484.0 -2.999, 0, -484.0 -2.999]),
            decimal=3
        )

    def test_orders_with_price_and_quantity(self):
        p = Portfolio()

        p.trade('MSFT', 'MSFT', '2021-01-01 00:00:00', 'B', None, 10, price=2.5, fee=-1)
        p.trade('MSFT', 'MSFT', '2021-01-01 00:00:00', 'B', None, -5, price=2, fee=-1)
        pf = p.get_current_portfolio()
        # print(pf)

        numpy.testing.assert_array_almost_equal(
            pf.loc['USD'].sum().values,
            np.array([-25 +10 -2, -25 +10 -2, 0, -25 +10 -2]),
            decimal=3
        )
        numpy.testing.assert_array_almost_equal(
            pf.loc["MSFT"].sum().values,
            np.array([5, 15, -2, -np.inf]),
            decimal=3
        )

    def test_orders_with_price_and_target_quantity(self):
        p = Portfolio()

        p.trade('MSFT', 'MSFT', '2021-01-01 00:00:00', 'B', None, TargetQuantity(10), price=2.5, fee=-1)
        p.trade('MSFT', 'MSFT', '2021-01-01 00:00:00', 'B', None, TargetQuantity(5), price=2, fee=-1)
        pf = p.get_current_portfolio()
        #print(pf)

        numpy.testing.assert_array_almost_equal(
            pf.loc['USD'].sum().values,
            np.array([-25 + 10 - 2, -25 + 10 - 2, 0, -25 + 10 - 2]),
            decimal=3
        )
        numpy.testing.assert_array_almost_equal(
            pf.loc["MSFT"].sum().values,
            np.array([5, 15, -2, -np.inf]),
            decimal=3
        )

    def test_orders_with_price_and_target_weight(self):
        p = Portfolio(capital=100)
        p.price.push_quote('MSFT', '2021-01-01 00:00:00', bid=10, ask=10)
        p.price.push_quote('AAPL', '2021-01-01 00:00:00', bid=0.1, ask=0.1)
        p.price.push_quote('VXX', '2021-01-01 00:00:00', bid=0.1, ask=0.1)
        p.price.push_quote('VXY', '2021-01-01 00:00:00', bid=0.1, ask=0.1)

        p.trade('MSFT', 'MSFT', '2021-01-02 00:00:00', 'B', None, TargetWeight(0.3), fee=0)
        p.trade('AAPL', 'AAPL', '2021-01-02 00:00:00', 'B', None, TargetWeight(0.7), fee=0)
        p.trade('VXX', 'VXX', '2021-01-02 00:00:00', 'B', None, TargetWeight(-0.2), fee=0)
        p.trade('VXY', 'VXY', '2021-01-02 00:00:00', 'B', None, TargetWeight(0.2), fee=0)

        pf = p.get_current_portfolio()
        #print(pf)

        numpy.testing.assert_array_almost_equal(
            pf["nav"].values,
            np.array([70, 30, 0, -20, 20]),
            decimal=3
        )

    def test_orders_with_price_and_target_weight_rebalance(self):
        # test bid/ask
        p = Portfolio(capital=100)
        p.price.push_quote('MSFT', '2021-01-01 00:00:00', bid=9.99, ask=10)
        p.trade('MSFT', 'MSFT', '2021-01-02 00:00:00', 'B', None, TargetWeight(0.3), fee=0)
        p.trade('MSFT', 'MSFT', '2021-01-02 00:00:00', 'B', None, TargetWeight(0.6), fee=0)
        pf = p.get_current_portfolio()
        print(pf)

        numpy.testing.assert_array_almost_equal(
            pf[["nav", "liquidation_value"]].values.T,
            np.array([[60.012, 39.988],
                      [59.951988, 39.988000]]),
            decimal=3
        )

    def test_with_quantity_no_price(self):
        p = Portfolio()

        #  price = 2.5
        p.price.push_quote('MSFT', '2021-01-02 00:00:00', bid=2, ask=2.5)
        p.trade('MSFT', 'MSFT', '2021-01-02 00:00:01', 'B', None, 10, fee=-1)

        p.price.push_quote('MSFT', '2021-01-03 00:00:00', bid=2, ask=2.5)
        p.trade('MSFT', 'MSFT', '2021-01-03 00:00:01', 'B', None, TargetQuantity(5), fee=-1)
        pf = p.get_current_portfolio()
        # print(pf)

        numpy.testing.assert_array_almost_equal(
            pf.loc['USD'].sum().values,
            np.array([-25 + 10 - 2, -25 + 10 - 2, 0, -25 + 10 - 2]),
            decimal=3
        )
        numpy.testing.assert_array_almost_equal(
            pf.loc["MSFT"].sum().values,
            np.array([5, 15, -2, 10]),
            decimal=3
        )

    def test_lala(self):
        p = Portfolio()

        for trade in self.get_trades():
            p.trade(*trade)

        pf = p.get_current_portfolio()
        print(pf)

        pf2 = p.foo(False)
        print(pf2)
        numpy.testing.assert_array_almost_equal(pf.values, pf2.values)

        pf3 = p.foo(True)
        print(pf3)

        numpy.testing.assert_array_almost_equal(p.get_portfolio_timeseries().values[:,:3], pf3.values[:,:3])


from typing import Callable


def ta_div(df, a: Callable, b: Callable, name=None):
    res = a(df) / b(df).values
    if name is not None:
        if res.ndims > 1:
            res.columns = name
        else:
            res.name = name

    return res


from unittest import TestCase

from pandas_ta_quant.technical_analysis import ta_decimal_year, ta_sinusoidal_week_day
from pandas_ta_quant_test.config import DF_TEST, DF_TEST_MULTI_ROW
import pandas as pd


class TestTimeEncoders(TestCase):

    def test_time_encoder(self):
        df = DF_TEST["Close"].copy()
        dec = ta_decimal_year(df)

        print(df.index[0], df.index[-1])
        print(dec[0], dec[-1])

        self.assertAlmostEqual(1993.0765027322404, dec[0])
        self.assertAlmostEqual(2019.9234972677596, dec[-1])

    def test_time_encoder_multi_index_row(self):
        df = DF_TEST_MULTI_ROW.copy()
        dec = ta_decimal_year(df)

        print(df.index[0], df.index[-1])
        print(dec[0], dec[-1])

        self.assertIsInstance(dec.index, pd.MultiIndex)
        self.assertAlmostEqual(1993.0765027322404, dec[0])
        self.assertAlmostEqual(2020.2349726775956, dec[-1])

    def test_sinusidal_week(self):
        df = DF_TEST_MULTI_ROW.copy()
        dec = ta_sinusoidal_week_day(df)

        print(df.index[0], df.index[-1])
        print(dec[0], dec[-1])

        self.assertAlmostEqual(-0.8660254037844384, dec[0])
        self.assertAlmostEqual(-0.8660254037844384, dec[-1])
        self.assertAlmostEqual(0, dec[-5])




from unittest import TestCase

import numpy as np
import pandas as pd

from pandas_ta_quant._decorators import for_each_top_level_column, is_time_consuming
from pandas_ta_quant._utils import conditional_func
from pandas_ta_quant.technical_analysis import ta_cci, ta_adx, ta_vola_hurst
from pandas_ta_quant_test.config import DF_TEST_MULTI


class TestAnalysisUtils(TestCase):

    def test_conditional_func(self):
        s = pd.Series(np.linspace(0, 1, 10))
        res = conditional_func(s >= 0.5, s * -10, s * 10)

        print(res)

    def test_decorators(self):
        """given a multi index data frame"""
        df = DF_TEST_MULTI

        """when executing an indicator"""
        func = for_each_top_level_column(ta_cci)
        res = func(df)

        func2 = for_each_top_level_column(ta_adx)
        res2 = func2(df)

        """then we got a column for each top level column"""
        self.assertEqual(2, res.columns.nlevels)
        self.assertEqual(2, res2.columns.nlevels)
        self.assertEqual(10, len(res2.columns))


    def test_has_decorator(self):
        self.assertTrue(_test_has_dec_true._is_timeconsuming)
        self.assertTrue(ta_vola_hurst._is_timeconsuming)
        self.assertFalse(hasattr(_test_has_dec_false, '_is_timeconsuming'))
        self.assertEqual(_test_has_dec_true(), 12)
        self.assertEqual(_test_has_dec_false(), 12)


def _test_decorator(func):
    def wrapper(*args, **kwargs):
        return func(*args, **kwargs)

    wrapper._is_decorated = True
    return wrapper


@is_time_consuming
def _test_has_dec_true():
    return 12


def _test_has_dec_false():
    return 12



"""Augment pandas DataFrame with methods for technical quant analysis"""
__version__ = '0.2.0'
import logging

from pandas.core.base import PandasObject

from pandas_ml_common import pd, np
from pandas_ta_quant.pandas_patch import TechnicalAnalysis as _TA
from pandas_ta_quant.portfolio import Portfolio

_log = logging.getLogger(__name__)
_log.debug(f"numpy version {np.__version__}")
_log.debug(f"pandas version {pd.__version__}")


setattr(PandasObject, "ta", property(lambda self: _TA(self)))



from unittest import TestCase

import pandas as pd

from pandas_ta_quant.technical_analysis import ta_draw_down
from pandas_ta_quant_test.config import DF_TEST

pd.set_option('display.max_columns', None)


class TestDrawDown(TestCase):

    def test_draw_down(self):
        df = DF_TEST.copy()
        dd = ta_draw_down(df["Close"], return_dates=True, return_duration=True)

        self.assertAlmostEqual(-0.035417, dd.loc["1993-03-02"]["mdd"], 6)
        self.assertListEqual(
            [0.000000, 0.000000, pd.Timestamp("1993-02-04"), pd.Timestamp("1993-03-03"), 18.0],
            dd.loc["1993-03-03"].to_list()
        )



from unittest import TestCase

import pandas as pd
import numpy as np

from pandas_ta_quant.technical_analysis import ta_candles_as_culb
from pandas_ta_quant_test.config import DF_TEST_MULTI


class TestMultiIndexColumn(TestCase):

    def test_culbs(self):
        df = DF_TEST_MULTI.copy()
        multi = ta_candles_as_culb(df)
        self.assertEqual(5*2, multi.shape[1])



from typing import Callable
from unittest import TestCase
from pandas_ta_quant import pd, np


class TestPatching(TestCase):

    def test_patched_pandas(self):
        df = pd.DataFrame({"a": range(10), "b": range(10)})
        self.assertIsInstance(df.ta.sma, Callable)

    def test_help_frame(self):
        df = pd.DataFrame({"a": range(10), "b": range(10)})
        help = df.ta.help

        self.assertGreater(len(help), 50)
        self.assertIn("doc", help)



from unittest import TestCase

from pandas_ta_quant.technical_analysis import ta_trend_lines
from pandas_ta_quant_test.config import DF_TEST


class TestSupportResistence(TestCase):

    def test_trend_lines(self):
        df = DF_TEST["Close"][-200:].copy()

        tl = ta_trend_lines(df)
        print(tl)


import numpy as _np
import pandas as _pd
import logging
from pandas_ml_common import get_pandas_object as _get_pandas_object
from pandas_ta_quant._decorators import *

_log = logging.getLogger(__name__)


@for_each_top_level_row
@for_each_top_level_column
def ta_realative_candles(df: _pd.DataFrame, open="Open", high="High", low="Low", close="Close", volume="Volume", drop_nan_volume=True):
    relative = _pd.DataFrame(index=df.index)
    relative[open] = (_np.log(df[open]) - _np.log(df[close].shift(1)))
    relative[close] = (_np.log(df[close]) - _np.log(df[close].shift(1)))
    relative[high] = (_np.log(df[high]) - _np.log(df[close].shift(1)))
    relative[low] = (_np.log(df[low]) - _np.log(df[close].shift(1)))

    if volume is not None:
        rel_vol = relative[volume].pct_change()
        if not drop_nan_volume or not rel_vol.isnull().all():
            relative[volume] = rel_vol

    return relative


@for_each_top_level_row
@for_each_top_level_column
def ta_candles_as_culb(df: _pd.DataFrame, open="Open", high="High", low="Low", close="Close", volume="Volume", relative_close=False, drop_nan_volume=True):
    o = _get_pandas_object(df, open)
    c = _get_pandas_object(df, close)
    h = _get_pandas_object(df, high)
    l = _get_pandas_object(df, low)
    oc = _pd.concat([o, c], axis=1)
    c_1 = c.shift(1)

    if not c.empty:
        # calculate close, upper_shadow, lower_shadow, body
        res = _pd.DataFrame({
            "close": (c / c_1 - 1) if relative_close else c,
            "upper": (h / oc.max(axis=1) - 1),
            "lower": (oc.min(axis=1) / l - 1),
            "body": (c / o - 1)
        }, index=df.index)
    else:
        _log.warning("empty DataFrame!")
        res = _pd.DataFrame({
            "close": [],
            "upper": [],
            "lower": [],
            "body": []
        })

    if volume is not None:
        rel_vol = df[volume].pct_change()
        if not drop_nan_volume or not rel_vol.isnull().all():
            res[volume] = rel_vol

    return res


@for_each_top_level_row
@for_each_top_level_column
def ta_candle_category(df: _pd.DataFrame, open="Open", high="High", low="Low", close="Close", body_threshold=0.975, gap_thresold=0.002):
    """
    We try to classify single candle sticks based on simple rules:
        t = treshold i.e. 975%

        red, green candles:
          g: positive candle
          r: negative candle

        opening gaps
          positive opening gap (x3, x1)
          negative opening gap (x1, x3)
          no gap (x2 / x2)

        shapes:
            https://blog.quantinsti.com/candlestick-patterns-meaning/
            1. body 99 % of hl
            2. lower shadow > upper shadow && shadows > body
            3. body > upper + lower shadow
            4. shadows > body
            5. upper shadow > lower shadow  && shadows > body

    :param df: data frame containing the open, high, low, close data
    :return: integer number of category: 2 * 3 * 5 = 30
    """

    df = df.copy()
    df["close-1"] = df[close].shift(1)

    def map(df):
        o, h, l, c, c1 = df[open], df[high], df[low], df[close], df["close-1"]
        hl = (h / l - 1)
        gap = o / c1 - 1
        sign, body_upper, body_lower, rank = (1, c, o, [5, 4, 1, 2, 3]) if c > o else (-1, o, c, [-5, -1, -4, -2, -3])
        body = body_upper / body_lower - 1
        lower_shadow = body_lower / l - 1
        upper_shadow = h / body_upper - 1
        shadow = lower_shadow + upper_shadow

        if abs(gap) > gap_thresold:
            if sign > 0:
                gap = 3 if gap > 0 else 1
            else:
                gap = 3 if gap < 0 else 1
        else:
            gap = 2

        if body >= (hl * body_threshold):
            return rank[0] * gap
        else:
            if shadow > body:
                if lower_shadow > (upper_shadow + body):
                    return rank[1] * gap
                elif upper_shadow > (lower_shadow + body):
                    return rank[2] * gap
                else:
                    return rank[3] * gap
            else:
                return rank[4] * gap

    ranks = df.apply(map, raw=False, axis=1)
    return ranks.rename("candle_type")



from typing import Union as _Union, Iterable

import numpy as _np
import pandas as _pd

from pandas_ta_quant._decorators import *

_PANDAS = _Union[_pd.DataFrame, _pd.Series]


@for_each_top_level_row
@for_each_column
def ta_std_ret_bands(s: _pd.Series, period=12, stddevs=[2.0], ddof=1, lag=1, scale_lag=True, include_mean=True, inculde_width=True) -> _PANDAS:
    if not isinstance(stddevs, Iterable):
        stddevs = [stddevs]

    mean = s.rolling(period).mean()
    std = s.pct_change().rolling(period).std(ddof=ddof).shift(lag) * (_np.sqrt(lag) if scale_lag else 1)
    res = mean.to_frame() if include_mean else _pd.DataFrame({}, index=s.index)

    for z in stddevs:
        res[f"upper_{z}"] = mean * (1 + std * z)
        res[f"lower_{z}"] = mean * (1 - std * z)

    if inculde_width:
        res["width"] = _np.log(res[f"upper_{stddevs[-1]}"] / res[f"lower_{stddevs[-1]}"]).rename("width")

    return res


@for_each_top_level_row
@for_each_column
def ta_bbands(df: _PANDAS, period=12, stddev=2.0, ddof=1, include_mean=True) -> _PANDAS:
    mean = df.rolling(period).mean().rename("mean")
    std = df.rolling(period).std(ddof=ddof)
    most_recent = df.rolling(period).apply(lambda x: x[-1], raw=True)

    upper = (mean + (std * stddev)).rename("upper")
    lower = (mean - (std * stddev)).rename("lower")
    z_score = ((most_recent - mean) / std).rename("z-score")
    width = _np.log(upper / lower).rename("width")
    quantile = _pd.Series(None, index=df.index, name='quantile', dtype='float64')
    quantile[most_recent <= lower] = 0
    quantile[most_recent > lower] = 1
    quantile[most_recent > mean] = 2
    quantile[most_recent > upper] = 3

    if include_mean:
        return _pd.DataFrame(upper) \
            .join(mean) \
            .join(lower) \
            .join(z_score) \
            .join(quantile) \
            .join(width) \
            .sort_index(axis=1)
    else:
        return _pd.DataFrame(upper) \
            .join(lower) \
            .join(z_score) \
            .join(quantile) \
            .join(width) \
            .sort_index(axis=1)



from typing import Union as _Union

# create convenient type hint
import numpy as _np
import numpy as np
import pandas as _pd
import pandas as pd

import pandas_ta_quant.technical_analysis.filters as _f
from pandas_ml_common import get_pandas_object as _get_pandas_object
from pandas_ta_quant._decorators import *

_PANDAS = _Union[_pd.DataFrame, _pd.Series]


@for_each_top_level_row
@for_each_top_level_column
def ta_tr(df: _PANDAS, high="High", low="Low", close="Close", relative=True) -> _PANDAS:
    h = _get_pandas_object(df, high)
    l = _get_pandas_object(df, low)
    c = _get_pandas_object(df, close).shift(1)

    if relative:
        ranges = (h / l - 1).rename("a").to_frame() \
            .join((h / c - 1).rename("b").abs()) \
            .join((l / c - 1).rename("c").abs())
    else:
        ranges = (h - l).rename("a").to_frame()\
            .join((h - c).rename("b").abs())\
            .join((l - c).rename("c").abs())

    return ranges.max(axis=1).rename("true_range")


@for_each_top_level_row
@for_each_top_level_column
def ta_atr(df: _PANDAS, period=14, high="High", low="Low", close="Close", relative=True, exponential='wilder') -> _PANDAS:
    if exponential is True:
        atr = _f.ta_ema(ta_tr(df, high, low, close, relative), period)
    if exponential == 'wilder':
        atr = _f.ta_wilders(ta_tr(df, high, low, close, relative), period)
    else:
        atr = _f.ta_sma(ta_tr(df, high, low, close, relative), period)

    return atr.rename(f"atr_{period}")


@for_each_top_level_row
@for_each_top_level_column
def ta_adx(df: _PANDAS, period=14, high="High", low="Low", close="Close", relative=True) -> _PANDAS:
    from pandas_ta_quant._utils import difference
    h = _get_pandas_object(df, high)
    l = _get_pandas_object(df, low)

    temp = _pd.DataFrame({
        "up": difference(h, h.shift(1), relative, 0),
        "down": difference(l.shift(1), l, relative, 0)
    }, index=df.index)

    atr = ta_atr(df, period, high, low, close, relative=relative)
    pdm = _f.ta_wilders(temp.apply(lambda r: r[0] if r["up"] > r["down"] and r["up"] > 0 else 0, raw=False, axis=1), period)
    ndm = _f.ta_wilders(temp.apply(lambda r: r[1] if r["down"] > r["up"] and r["down"] > 0 else 0, raw=False, axis=1), period)

    pdi = pdm / atr
    ndi = ndm / atr
    adx = _f.ta_wilders((pdi - ndi).abs() / (pdi + ndi).abs(), period)

    return _pd.DataFrame({"+DM": pdm, "-DM": ndm, "+DI": pdi, "-DI": ndi, "ADX": adx}, index=df.index)


@for_each_top_level_row
@for_each_top_level_column
def ta_williams_R(df: _pd.DataFrame, period=14, close="Close", high="High", low="Low") -> _pd.Series:
    temp = _get_pandas_object(df, close).to_frame()
    temp = temp.join(_get_pandas_object(df, high if high is not None else close).rolling(period).max().rename("highest_high"))
    temp = temp.join(_get_pandas_object(df, low if low is not None else close).rolling(period).min().rename("lowest_low"))
    return ((temp["highest_high"] - temp[close]) / (temp["highest_high"] - temp["lowest_low"])).rename(f"williams_R_{period}")


@for_each_top_level_row
@for_each_top_level_column
def ta_ultimate_osc(df: _pd.DataFrame, period1=7, period2=14, period3=28, close="Close", high="High", low="Low") -> _pd.Series:
    # BP = Close - Minimum(Low or Prior Close).
    # TR = Maximum(High or Prior Close)  -  Minimum(Low or Prior Close)
    prev_close = _get_pandas_object(df, close).shift(1)
    downs = (_get_pandas_object(df, low if low is not None else close).to_frame().join(prev_close)).min(axis=1)
    ups = (_get_pandas_object(df, high if high is not None else close).to_frame().join(prev_close)).max(axis=1)
    temp = _pd.DataFrame({
        "bp": _get_pandas_object(df, close) - downs,
        "tr": ups - downs
    }, index=df.index)

    periods = [period1, period2, period3]
    avs = []
    for period in periods:
        # Average7 = (7 - period BP Sum) / (7 - period TR Sum)
        av = temp.rolling(period).sum()
        avs.append(av["bp"] / av["tr"])

    # UO = [(4 x Average7) + (2 x Average14) + Average28] / (4 + 2 + 1)
    return ((4 * avs[0] + 2 * avs[1] + avs[2]) / 7).rename(f"ultimate_osc_{period1},{period2},{period3}")


@for_each_top_level_row
@for_each_top_level_column
def ta_bop(df: _pd.DataFrame, open="Open", high="High", low="Low", close="Close") -> _PANDAS:
    # (CLOSE – OPEN) / (HIGH – LOW)
    o = _get_pandas_object(df, open)
    c = _get_pandas_object(df, close)
    h = _get_pandas_object(df, high)
    l = _get_pandas_object(df, low)

    return ((c - o) / (h - l)).rename("bop")


@for_each_top_level_row
@for_each_top_level_column
def ta_cci(df: _pd.DataFrame, period=14, high="High", low="Low", close="Close", alpha=0.015) -> _PANDAS:
    h = _get_pandas_object(df, high)
    l = _get_pandas_object(df, low)
    c = _get_pandas_object(df, close)

    tp = (h + l + c) / 3
    tp_sma = _f.ta_sma(tp, period)

    md = tp.rolling(period).apply(lambda x: _np.abs(x - x.mean()).sum() / period)
    return ((1 / alpha) * (tp - tp_sma) / md / 100).rename(f"cci_{period}")


@for_each_top_level_row
@for_each_top_level_column
def ta_gap(df: _pd.DataFrame, open="Open", close="Close") -> _PANDAS:
    return (df[open] / df[close].shift(1) - 1).rename("gap")


@for_each_top_level_row
@for_each_top_level_column
@is_time_consuming
def ta_vola_hurst(df: _PANDAS, period=255*2, lags=30, open="Open", high="High", low="Low", close="Close") -> _PANDAS:
    x = np.arange(1, lags)
    v = np.log(ta_gkyz_volatility(df, period=1, open=open, high=high, low=low, close=close)).rename("log_sqrt")

    def hurst(sig):
        # def del_Raw(s, q, x):
        #    return [np.mean(np.abs(s - s.shift(lag)) ** q) for lag in x]
        #
        # zeta_q = [np.polyfit(np.log(x), np.log(del_Raw(s, q, x)), 1)[0] for q in qVec]
        # h_est = np.polyfit(qVec, zeta_q, 1)[0]

        def dlsig2(sig, x):
            return [np.mean((sig - sig.shift(lag)) ** 2) for lag in x]

        model = np.polyfit(np.log(x), np.log(dlsig2(sig, x)), 1)
        return np.array([model[0] / 2., np.sqrt(np.exp(model[1]))])

    e_hurst = np.array([hurst(v.iloc[i-period+1:i]) for i in range(period - 1, len(df))])

    return df[[]].join(
        pd.DataFrame(e_hurst, columns=[f"H_{period}/{lags}", f"nu_{period}/{lags}"], index=df.index[period-1:]))


@for_each_top_level_row
@for_each_top_level_column
def ta_gkyz_volatility(df: _pd.DataFrame, period=12, open="Open", high="High", low="Low", close="Close") -> _PANDAS:
    # sqrt (sum of (  ln(o[i] / c[i-1] )^2 + 1/2 * (ln(h[i] / l[i]))^2 - (2 * ln(2) - 1) * (ln(c[i] / o[i]))^2   ) / N)
    vdf = df[[open, high, low, close]].copy()
    vdf.columns = ["o", "h", "l", "c"]
    vdf["c_1"] = df[close].shift(1)
    ln = np.log

    def one(r):
        return ln(r["o"] / r["c_1"])**2 + 0.5*(ln(r["h"] / r["l"]))**2 - (2 * ln(2) - 1)*(ln(r["c"] / r["o"]))**2

    v = np.sqrt(vdf.apply(one, axis=1).rolling(period).mean())
    return v.rename(f"gkyz_vol_{period}")


@for_each_top_level_row
@for_each_top_level_column
def ta_cc_volatility(df: _PANDAS, period=12, close="Close") -> _PANDAS:
    return np.sqrt((np.log(df[close] / df[close].shift(1))**2).rolling(period).mean()).rename(f"cc_vol_{period}")







from unittest import TestCase

from pandas_ta_quant import np
from pandas_ta_quant.technical_analysis.encoders import ta_candles_as_culb
from pandas_ta_quant_test.config import DF_TEST


class TestCandleEncoder(TestCase):

    def test_multi_channel(self):
        df = DF_TEST.copy()
        ohlc_values = df[["Open", "High", "Low", "Close"]].iloc[-1].values
        culb = ta_candles_as_culb(df, volume=None)
        culb_values = culb.iloc[-1].values

        self.assertGreaterEqual(culb["upper"].values.max(), 0)
        self.assertGreaterEqual(culb["lower"].values.max(), 0)
        self.assertEqual(ohlc_values[1] > ohlc_values[-1], culb_values[-1] < 0)
        np.testing.assert_array_almost_equal(np.array([3.120900e+02, 6.402011e-05, 4.861506e-03, -4.48435e-04]),
                                             culb_values, 5)




from unittest import TestCase

from pandas_ta_quant.technical_analysis import ta_multi_ma, ta_naive_edge_detect, ta_edge_detect_mean, \
    ta_edge_detect_poly
from pandas_ta_quant_test.config import DF_TEST
import numpy as np


class TestFilter(TestCase):

    def test_multi_ma(self):
        df = DF_TEST
        mma = ta_multi_ma(df)

        self.assertEqual(len(mma.columns), 30)

    def test_naive_edge(self):
        edges1 = ta_naive_edge_detect(DF_TEST["Close"])
        edges2 = ta_edge_detect_mean(DF_TEST["Close"])
        edges3 = ta_edge_detect_poly(DF_TEST["Close"])
        print(DF_TEST[["Close"]].join(edges1).join(edges2).join(edges3).tail(20))




from functools import wraps

import pandas as pd
import logging
from pandas_ml_common import unique_level_rows, unique_level_columns
from pandas_ml_common.utils import same_columns_after_level

_log = logging.getLogger(__name__)


def for_each_column(func):
    @wraps(func)
    def exec_on_each_column(df: pd.DataFrame, *args, **kwargs):
        if df.ndim > 1 and df.shape[1] > 0:
            results = [func(df[col], *args, **kwargs) for col in df.columns]
            if results[0].ndim > 1 and results[0].shape[1] > 1:
                for i, col in enumerate(df.columns):
                    results[i].columns = pd.MultiIndex.from_product([[col], results[i].columns.tolist()])

            return pd.concat(results, axis=1, join='inner')
        else:
            return func(df, *args, **kwargs)

    return exec_on_each_column


def for_each_top_level_column(func, level=0):
    @wraps(func)
    def exec_on_each_tl_column(df: pd.DataFrame, *args, **kwargs):
        if df.ndim > 1 and isinstance(df.columns, pd.MultiIndex):
            # check if the shape of the 2nd level is identical else threat as if not multi index
            if same_columns_after_level(df, level):
                top_level = unique_level_columns(df, level)
                groups = [func(df.xs(group, axis=1, level=level), *args, **kwargs).to_frame().add_multi_index(group, inplace=True, level=level) for group in top_level]
                return pd.concat(groups, axis=1)
            else:
                _log.warning(f"columns in further levels do not follow the same structure! Treat as normal Index")
                return func(df, *args, **kwargs)
        else:
            return func(df, *args, **kwargs)

    return exec_on_each_tl_column


def for_each_top_level_row(func):
    @wraps(func)
    def exec_on_each_tl_row(df: pd.DataFrame, *args, **kwargs):
        if isinstance(df.index, pd.MultiIndex):
            top_level = unique_level_rows(df, 0)
            if len(top_level) > 1:
                return pd.concat(
                    [func(df.loc[group], *args, **kwargs).add_multi_index(group, inplace=True, axis=0) for group in top_level],
                    axis=0
                )
            else:
                return func(df, *args, **kwargs)
        else:
            return func(df, *args, **kwargs)

    return exec_on_each_tl_row


def is_time_consuming(func):
    def wrapper(*args, **kwargs):
        return func(*args, **kwargs)

    wrapper._is_timeconsuming = True
    return wrapper



import numpy as np
import pandas as pd
from numba import guvectorize, float32, int32, float64, int64

from pandas_ml_common.utils import ReScaler


def sort_distance(s1: pd.Series, s2: pd.Series, top_percent=None) -> np.ndarray:
    s = (s1 / s2 - 1).dropna().abs()
    distances = s.sort_values().values

    if top_percent is not None:
        return distances[:int(np.ceil(len(distances) * top_percent))]
    else:
        return distances


def symmetric_quantile_factors(a, bins=11):
    return np.linspace(1-a, 1+a, bins)


def conditional_func(s: pd.Series, s_true: pd.Series, s_false: pd.Series):
    df = s.to_frame("CONDITION").join(s_true.rename("TRUTHY")).join(s_false.rename("FALSY"))
    return df.apply(lambda r: r["TRUTHY"] if r["CONDITION"] is True else r["FALSY"] , axis=1)


def difference(a: pd.Series, b: pd.Series, relative: bool, replace_inf=0):
    if relative:
        if replace_inf is not None:
            return (a / b - 1).replace([np.inf, -np.inf], replace_inf)
        else:
            return a / b - 1
    else:
        return a - b


def returns_to_log_returns(returns):
    return np.log(1 + returns)


def log_returns_to_returns(log_returns):
    return (np.e ** log_returns) - 1


@guvectorize([(float32[:], int32, float32[:]),
              (float64[:], int64, float64[:])], '(n),()->(n)')
def wilders_smoothing(arr: np.ndarray, period: int, res: np.ndarray):
    assert period > 0
    alpha = (period - 1) / period
    beta = 1 / period

    res[0:period] = np.nan
    res[period - 1] = arr[0:period].mean()
    for i in range(period, len(arr)):
        res[i] = alpha * res[i-1] + arr[i] * beta


def _rescale(df: pd.DataFrame, range=(-1, 1), digits=None, axis=None):
    if axis is not None:
        return df.apply(lambda x: _rescale(x, range), raw=False, axis=axis, result_type='broadcast')
    else:
        rescaler = ReScaler((df.values.min(), df.values.max()), range)
        rescaled = rescaler(df.values)

        if digits is not None:
            rescaled = np.around(rescaled, digits)

        if rescaled.ndim > 1:
            return pd.DataFrame(rescaled, columns=df.columns, index=df.index)
        else:
            return pd.Series(rescaled, name=df.name, index=df.index)



# TODO eventually turn this into a decorator ???
def with_column_suffix(suffix, po, ref_po=None):
    if ref_po is None:
        ref_po = po

    if po.ndim > 1:
        if isinstance(po.columns, pd.MultiIndex):
            po.columns = pd.MultiIndex.from_tuples([(f'{col[0]}_{suffix}', *col[1:]) for col in ref_po.columns.to_list()])
            return po
        else:
            po.columns = ref_po.columns
            return po.add_suffix(f'_{suffix}')
    else:
        if isinstance(po.name, tuple):
            return po.rename((suffix, *ref_po.name))
        else:
            return po.rename(f'{ref_po.name}_{suffix}')


# return index of bucket of which the future price lies in
def index_of_bucket(value, data):
    if np.isnan(value) or np.isnan(data).any() or np.isinf(value) or np.isinf(value).any():
        return np.nan

    for i, v in enumerate(data):
        if value < v:
            return i

    return len(data)


def has_decorator(function, decorator):
    # If we have no func_closure, it means we are not wrapping any other functions.
    if not function.func_closure:
        return False

    for closure in function.func_closure:
        if has_decorator(closure.cell_contents):
            return True

    return [function]


def rolling_apply(df, period, func, names, center=False, **kwargs):
    assert center is False or period % 2 > 0, "only odd periods are allowed"
    margin = (period - 1) // 2 if center else 0
    index = df.index[period - margin - 1: len(df) - margin]

    def fi(i):
        return i - period + margin

    def ti(i):
        return i + margin

    # if period is 3 dann [0:3]
    data = np.array([func(df.iloc[fi(i):ti(i)], **kwargs) for i in range(period-margin, len(df)+1-margin)])
    s = pd.Series(data, index=index, name=names) if data.ndim < 2 else pd.DataFrame(data, index=index, columns=names)

    if df.ndim < 2:
        df = df.to_frame()

    return df[[]].join(s)


# create convenient type hint

import numpy as np
import pandas as pd

from pandas_ml_common import Typing as _t
from pandas_ta_quant._decorators import *
from pandas_ta_quant._utils import rolling_apply


@for_each_top_level_row
@for_each_column
def ta_naive_edge_detect(df: _t.PatchedDataFrame, period=3):
    assert period % 2 > 0, "only odd periods are allowed"
    center = (period - 1) // 2

    return df.rolling(period, center=True)\
             .apply(lambda s: 1 if (s[0] < s[center] > s[-1]) else -1 if (s[0] > s[center] < s[-1]) else 0, raw=True)\
             .rename(f'{df.name}_edge_naive_{period}')


@for_each_top_level_row
@for_each_column
def ta_edge_detect_mean(df: _t.PatchedSeries, period=3):
    assert df.ndim == 1 or len(df.columns) == 1, "Trend lines can only be calculated on a series"
    assert period > 2, "minimum period is 3"

    def edge(col):
        mean = col.mean()
        if col[0] > mean and col[-1] > mean:
            return 1
        elif col[0] < mean and col[-1] < mean:
            return -1
        else:
            return 0

    return df.rolling(period, center=True).apply(edge, raw=True).rename(f'{df.name}_edge_mean_{period}')


@for_each_top_level_row
@for_each_column
def ta_edge_detect_poly(df: _t.PatchedSeries, period=3, a_threshold=5, vx_threshold=0.1):
    x = np.linspace(0, 1, period)
    ab = rolling_apply(df, period, lambda s: np.polyfit(x, s.values, deg=2)[:2], ["a", "b"], center=True)

    # x of vertex will be -b/2a
    ab["vx"] = ab.apply(lambda r: -r["b"] / (2 * r["a"]), axis=1)

    if a_threshold is not None:
        ab["a"] = ab["a"].apply(lambda x: x if np.abs(x) >= a_threshold else 0)

    if vx_threshold is not None:
        # a perfect turning point would be at x=1/2
        ab["a"] = ab[["a", "vx"]].apply(lambda x: x["a"] if np.abs(x["vx"] - 0.5) <= vx_threshold else 0, axis=1)

    return df.to_frame()[[]].join(ab["a"].rename(f'{df.name}_edge_poly_{period}'))


EDGE_DETECTOR = {
    "mean": ta_edge_detect_mean,
    "naive": ta_naive_edge_detect,
    "poly": ta_edge_detect_poly,
}



import os

import numpy as np

from pandas_ta_quant import pd

print('NUMPY VERSION', np.__version__)


DF_TEST_MULTI = pd.read_pickle(os.path.join(os.path.dirname(os.path.abspath(__file__)), ".data", "spy_gld.pickle"))
DF_TEST = pd.read_csv(os.path.join(os.path.dirname(os.path.abspath(__file__)), ".data", "SPY.csv"), index_col='Date', parse_dates=True)
DF_TEST2 = pd.read_csv(os.path.join(os.path.dirname(os.path.abspath(__file__)), ".data", "GLD.csv"), index_col='Date', parse_dates=True)
DF_TEST_MULTI_ROW = pd.concat([DF_TEST.add_multi_index("SPY", axis=0), DF_TEST2.add_multi_index("GLD", axis=0)], axis=0)
DF_TEST_MULTI_ROW_MULTI_COLUMN = pd.concat([DF_TEST_MULTI.add_multi_index("A", axis=0), DF_TEST_MULTI.add_multi_index("B", axis=0)], axis=0)
DF_DEBUG = pd.DataFrame({"Close": np.random.random(10)})

DF_INVERSE_GAF = pd.read_pickle(os.path.join(os.path.dirname(os.path.abspath(__file__)), ".data", "inverse_gaf.df"))
DF_TEST_MULTI_CLASS = pd.read_pickle(os.path.join(os.path.dirname(os.path.abspath(__file__)), ".data", "one_hot_classified_df.pickle"))

CSV_TRADES = os.path.join(os.path.dirname(os.path.abspath(__file__)), ".data", "trades.csv")




import numpy as np

from pandas_ml_common import Typing, np_nans
from pandas_ta_quant._decorators import *
from pandas_ta_quant._utils import with_column_suffix as _wcs, _rescale
from pandas_ta_quant.technical_analysis import filters as _f


@for_each_top_level_row
def ta_rescale(df: pd.DataFrame, range=(-1, 1), digits=None, axis=None):
    return _rescale(df, range, digits, axis)


@for_each_top_level_row
def ta_performance(df: Typing.PatchedPandas):
    return _wcs("performance", (1 + df.pct_change()).cumprod())


@for_each_top_level_row
def ta_returns(df: Typing.PatchedPandas, period=1):
    return _wcs("return", df.pct_change(periods=period))


@for_each_top_level_row
def ta_cumret(df: Typing.PatchedPandas, period=1):
    if period == 1:
        return (1 + df).cumprod()
    else:
        rets = np_nans((len(df), df.shape[1] if df.ndim > 1 else 1))
        rets[:period] = 1
        for i in range(period, len(df), period):
            rets[i] = (df.iloc[i] + 1) * rets[i-period]

        if df.ndim > 1:
            return pd.DataFrame(rets, columns=df.columns, index=df.index)
        else:
            return pd.Series(rets.squeeze(), name=df.name, index=df.index)


@for_each_top_level_row
def ta_log_returns(df: Typing.PatchedPandas, period=1):
    current = df
    lagged = df.shift(period)

    return _wcs("log_return", np.log(current) - np.log(lagged))


def ta_logret_as_return(df: Typing.PatchedPandas):
    return np.exp(df)


@for_each_top_level_row
def ta_cumlogret(df: Typing.PatchedPandas, period=1):
    return ta_cumret(ta_logret_as_return(df) - 1, period)


@for_each_top_level_row
def ta_ma_ratio(df: Typing.PatchedPandas, period=12, lag=0, ma='sma', **kwargs):
    mafunc = getattr(_f, f'ta_{ma}')
    return _wcs(f"{ma}({period}) x 1/", df / mafunc(df, period=period, **kwargs).shift(lag).values - 1, df)

# TODO reconstruct sma ratios


import numpy as _np
import pandas as _pd

# create convenient type hint
from pandas_ml_common import Typing as _t
from pandas_ta_quant._decorators import *


@for_each_top_level_row
def ta_decimal_year(df: _t.PatchedPandas):
    return ((df.index.strftime("%j").astype(float) - 1) / 366 + df.index.strftime("%Y").astype(float))\
        .to_series(index=df.index, name="decimal_time")


@for_each_top_level_row
def ta_sinusoidal_week_day(po: _t.PatchedPandas):
    if not isinstance(po.index, _pd.DatetimeIndex):
        df = po.copy()
        df.index = _pd.to_datetime(df.index)
    else:
        df = po

    return _np.sin(2 * _np.pi * (df.index.dayofweek / 6.0)).to_series(index=po.index, name="dow")


@for_each_top_level_row
def ta_sinusoidal_week(po: _t.PatchedPandas):
    if not isinstance(po.index, _pd.DatetimeIndex):
        df = po.copy()
        df.index = _pd.to_datetime(df.index)
    else:
        df = po

    return _np.sin(2 * _np.pi * (df.index.isocalendar().week / 52.0)).rename("week")



from unittest import TestCase

from pandas_ta_quant.technical_analysis import ta_sarimax, ta_sma
from pandas_ta_quant_test.config import DF_TEST


class TestForecast(TestCase):

    def test_sarimax(self):
        df = DF_TEST[-65:]

        df = ta_sarimax(ta_sma(df["Close"], 2).pct_change().dropna())
        print(df)
        self.assertAlmostEqual(0.0014875, df.iloc[-1, -1], 6)





import numbers
import uuid
from collections import namedtuple, defaultdict
from io import StringIO
from typing import Tuple, Union, Callable, List
import logging

import numpy as np
import pandas as pd

from pandas_ml_common.utils.time_utils import parse_timestamp, min_timestamp
from .price import PriceTimeSeries, AbstractPriceTimeSeries

_VAL_BASE_CURR = 'liquidation_value'
_POS_AVG_PRICE = 'nav'
TrxKey = namedtuple('TrxKey', ['underlying', 'instrument', 'timestamp'])
Order = namedtuple('Order', ['ID', 'order_type', 'order_subtype', 'currency', 'quantity', 'nav', 'fee', 'strategy'])
_log = logging.getLogger(__name__)


class Quantity(object):

    def __init__(self, v: float):
        self.v = v

    def value(self,
              lazy_current_portfolio: Callable[[str, str], pd.DataFrame],
              underlying: str,
              instrument: str,
              currency: str,
              prices: AbstractPriceTimeSeries,
              tst: pd.Timestamp,
              price: float = None,
              fee: float = 0,
             ):
        return self.v


class TargetQuantity(Quantity):

    def __init__(self, v: float):
        super().__init__(v)

    def value(self,
              lazy_current_portfolio: Callable[[str, str], pd.DataFrame],
              underlying: str,
              instrument: str,
              currency: str,
              prices: AbstractPriceTimeSeries,
              tst: pd.Timestamp,
              price: float = None,
              fee: float = 0,
             ):
        current_position = lazy_current_portfolio(underlying, instrument, currency)
        if current_position.empty:
            return self.v
        else:
            assert current_position.ndim <= 1, f'Expected Series, got DataFrame\n{current_position}'
            current_quantity = current_position['quantity'].item()
            quantity = self.v - current_quantity
            return quantity


class TargetWeight(Quantity):

    def __init__(self, v: float):
        assert -1 <= v <= 1, 'Weighs need to be between -1 and 1'
        super().__init__(v)
        self.warned_fee = False

    def value(self,
              lazy_current_portfolio: Callable[[str, str], pd.DataFrame],
              underlying: str,
              instrument: str,
              currency: str,
              prices: AbstractPriceTimeSeries,
              tst: pd.Timestamp,
              price: float = None,
              fee: float = 0,
             ):
        if fee != 0:
            if not self.warned_fee:
                _log.warning("we can't handle fees properly using TargetWeight. "
                             "Your PriceTimeSeries should use bid/ask or spread!")
                self.warned_fee = True

        current_portfolio = lazy_current_portfolio(None, None, None)
        try:
            current_cash_position = current_portfolio.loc[currency, currency][_POS_AVG_PRICE].sum()
        except KeyError:
            raise ValueError('Portfolio needs to be funded `Portfolio(capital=100)`')

        current_positions = current_portfolio.drop([currency, currency])[[_VAL_BASE_CURR]].sum().item()
        target_nav = (current_cash_position + current_positions) * self.v

        idx = (underlying, instrument)

        current_nav = current_portfolio.loc[idx][_VAL_BASE_CURR].sum().item() if idx in current_portfolio.index else 0
        if current_nav <= -np.inf:
            raise ValueError(f"TargetWeight needs prices! Missing price for {idx}")

        if price is not None:
            qty = (target_nav - current_nav) / price
        else:
            _, (bid, ask) = prices.get_price(instrument, tst, currency)
            price = bid if target_nav < current_nav else ask
            qty = (target_nav - current_nav) / price

        return qty


class Portfolio(object):

    @staticmethod
    def from_tradelog_file(file: Union[str, List[str]], *args, strategy_mapper: Callable[[TrxKey], str] = None, **kwargs):
        trandelog_files = [file] if not isinstance(file, (list, tuple)) else file
        content = defaultdict(list)

        for tradelog_file in trandelog_files:
            for line in open(tradelog_file).readlines():
                if len(line) <= 1:
                    continue

                if '|' not in line:
                    key = line.replace('\n', '')
                    continue

                content[key].append(line)

        log = {k: pd.read_csv(StringIO('\n'.join(v)), sep='|', parse_dates=True, header=None) for k, v in content.items()}
        transactions = {k: v for k, v in log.items() if k.endswith('_TRANSACTIONS')}

        pf = Portfolio(*args, **kwargs)
        for asset, trx_asset in transactions.items():
            for i, trx in trx_asset.iterrows():
                # transaction key
                if asset.startswith('STOCK'):
                    trx_key = TrxKey(trx[2], trx[2], pd.to_datetime(f'{trx[7]} {trx[8]}', format='%Y%m%d %H:%M:%S'))
                elif asset.startswith('OPTION'):
                    trx_key = TrxKey(trx[2].split(' ')[0], trx[3], pd.to_datetime(f'{trx[7]} {trx[8]}', format='%Y%m%d %H:%M:%S'))
                elif asset.startswith('FUTURE'):
                    trx_key = TrxKey(trx[3].split(' ')[0], trx[3], pd.to_datetime(f'{trx[7]} {trx[8]}', format='%Y%m%d %H:%M:%S'))
                else:
                    _log.warning(f'Not implemented {asset}')
                    continue

                # order type and sub type
                if trx[6] in ['A;C']:
                    ot = list(reversed(trx[6].split(';')))
                else:
                    ot = trx[6].split(';') + [None]

                strategy = None if strategy_mapper is None else strategy_mapper(trx_key)
                order = Order(trx[1], ot[0], ot[1], trx[9], trx[10], trx[13], trx[14], strategy)

                pf.trade(trx_key.underlying, trx_key.instrument, trx_key.timestamp,
                         order.order_type, order.order_subtype, order.quantity, order.nav, fee=order.fee,
                         strategy=order.strategy)

        return pf

    def __init__(self, prices: AbstractPriceTimeSeries = None, capital: float = None, currency: str = 'USD'):
        self.prices = PriceTimeSeries() if prices is None else prices
        self.currency = currency
        self.orders = []

        if capital is not None:
            self.trade(currency, currency, min_timestamp(), 'F', None, nav=capital, price=1, currency=currency, id=str(uuid.uuid4()))

    def trade(self,
              underlying: str,
              instrument: str,
              timestamp: Union[pd.Timestamp, str, Tuple[str, str]],
              order_type: str,
              order_subtype: str = None,
              quantity: Union[float, Quantity] = None,
              nav: float = None,
              price: float = None,
              currency: str = None,
              strategy: str = None,
              id: str = str(uuid.uuid4()),
              fee: float = 0,
             ):
        """
        :param underlying: any string
        :param instrument: any string, can be None we then just use underlying
        :param timestamp: any timestamp - NOTE it is important to pass a time (in UTC)!
            if you want to buy at "close" use 23:59:59.9999
        :param order_type: BO (Buy to Open), SC (Sell to Close), SO (Sell to Open), BC (Buy to Close), A (Assigned)
        :param order_subtype: None, E (Expired), O (Open), A (Assigned)
        :param quantity: any float
        :param nav: None or any float
        :param price: None,
            if price is given and nav is None, we calculate the nav using quantity * price
            if price is None and nav is None, we need a price lookup table and a lookup rule i.e. bid/ask, open/close, ...
                if no price lookup table provided throw exception
                but the whole API can be lazy such that we need prices only on get_trades_dataframe()
        :param fee: 0, or any flot or callable(nav, quantity)
            if fee is callable we calculate the fee using the qunatinty and nav value
        :param currency: USD, or any string
        :param strategy: None, or any sting (used as an analytics field)
        :param id: UUID or any other string/primitive

        :return:
        """
        assert nav is not None or quantity is not None, "Quantity or Net Asset Value (nav) need to be passed!"
        assert currency is None or currency == self.currency, "Multiple currencies are not supported at the moment"

        # parse timestamp
        tst = parse_timestamp(timestamp)

        # assign default currency
        if currency is None:
            currency = self.currency

        # calculate net asset value
        if nav is None:
            # calculate quantity
            if isinstance(quantity, numbers.Number):
                quantity = Quantity(quantity)

            quantity = quantity.value(lambda u, i, c: self.get_current_position(u, i, c),
                                      underlying, instrument, self.currency, self.prices, tst, price, fee)

            if price is None:
                price_tst, (bid, ask) = self.prices.get_price(instrument, tst, currency)
                price = bid if quantity < 0 else ask

            # we calculate the nav using quantity * price
            nav = quantity * price
        elif quantity is None:
            # calculate quantity
            if price is None:
                price_tst, (bid, ask) = self.prices.get_price(instrument, tst, currency)
                price = bid if nav < 0 else ask

            quantity = nav / price

        order = Order(id, order_type, order_subtype, currency, quantity, nav, fee, strategy)
        trx_key = TrxKey(underlying, underlying if instrument is None else instrument, tst)
        self.orders.append(pd.Series(order, index=order._fields, name=trx_key))

        if instrument != currency:
            # now also make a cash trade
            #  TODO later when we support multiple currencies also make an FX-Trade the very same way
            #   Note that we can not aggregate over different currencies anymore
            cash_ot = 'P' if nav > 0 else 'R'
            cash_id = f'{id}_{currency}'
            self.trade(currency, currency, timestamp, cash_ot, 'T', nav=-nav+fee, price=1, currency=currency, strategy=strategy, id=cash_id)

    @property
    def price(self):
        return self.prices

    def get_current_portfolio(self, strategy=None):
        return self.get_current_position(None, None, strategy)

    def get_current_position(self, underlying, instrument=None, currency=None, strategy=None):
        df_trades = self.get_order_sequence().drop(['order_type', 'order_subtype', 'strategy'], axis=1)
        agg = get_position_aggregator(self.prices, instrument, self.currency)

        if strategy is not None:
            df_trades = df_trades[df_trades["strategy"] == strategy]

        if underlying is None:
            underlying = slice(None)
        if instrument is None:
            instrument = slice(None)
        if currency is None:
            currency = self.currency

        res = df_trades.groupby(level=[0, 1]).apply(agg)
        return res if res.empty else res.loc[(underlying, instrument, currency)]

    def get_portfolio_timeseries(self):
        df_trades = self.get_order_sequence()
        return df_trades.groupby(level=[0, 1]).apply(get_position_timeseries_aggregator(self.prices, self.currency))

    def get_order_sequence(self) -> pd.DataFrame:
        df_trades = pd.DataFrame(self.orders, columns=Order._fields)
        df_trades.index = pd.MultiIndex.from_tuples(df_trades.index, names=TrxKey._fields)
        df_trades = df_trades.drop(['ID'], axis=1).sort_index()
        return df_trades

    def foo(self, ts):
        df_trades = self.get_order_sequence()
        if ts:
            return df_trades.groupby(level=[0, 1]).apply(get_position_timeseries_aggregator(self.prices, self.currency))
        else:
            return df_trades.groupby(level=[0, 1]).apply(get_position_aggregator(self.prices, None, self.currency))


def get_position_aggregator(prices: AbstractPriceTimeSeries, instrument: str, base_currency: str):
    def aggregator(pos: pd.Series):
        res = pos[['quantity', 'nav', 'fee', 'currency']].groupby('currency')
        res = res.sum()

        tst = pos.index.get_level_values(-1).max()
        inst = pos.name[1] if instrument is None else instrument

        def mark_to_market(pos):
            try:
                val = res['nav'] if inst == base_currency else \
                    prices.get_price(inst, tst, pos.name, base_currency)[1][0] * pos["quantity"].item()
            except KeyError:
                val = -np.inf
                # _log.error(f"Failed to get price for {inst} @ {tst}")

            return val

        res[_VAL_BASE_CURR] = res.apply(mark_to_market, axis=1)
        res = res.rename(columns={"nav": _POS_AVG_PRICE}) if res.ndim > 1 else res.rename(index={"nav": _POS_AVG_PRICE})
        return res

    return aggregator


def get_position_timeseries_aggregator(prices: AbstractPriceTimeSeries, base_currency: str):
    def aggregator(pos: pd.Series):
        grp = pos[['quantity', 'nav', 'fee', 'currency']].groupby('currency')
        grp = {g: f.drop('currency', axis=1).cumsum() for g, f in grp}
        res = pd.concat(grp.values(), axis=0, keys=grp.keys())

        def mark_to_market(pos):
            try:
                val = pos['nav'] if pos.name[1] == base_currency else \
                    prices.get_price(pos.name[1], pos.name[-1], pos["currency"].item(), base_currency)[1][0] * pos["quantity"].item()
            except KeyError:
                val = -np.inf
                # _log.error(f"Failed to get price for {inst} @ {tst}")

            return val

        res[_VAL_BASE_CURR] = res.apply(mark_to_market, axis=1)
        res.index = pd.MultiIndex.from_arrays([res.index.get_level_values(-1), res.index.get_level_values(0)])
        res.rename(columns={"nav": _POS_AVG_PRICE})

        return res

    return aggregator


from arch import arch_model  # alternatively https://pyflux.readthedocs.io/en/latest/garch.html

import pandas_ta_quant.technical_analysis.normalizer as _norm
from pandas_ml_common import Typing
from pandas_ta_quant._decorators import *
from functools import partial
import scipy.special as scsp
import numpy as np


@for_each_top_level_row
@for_each_column
def ta_garch11(df: Typing.PatchedPandas, period=200, forecast=1, returns='returns'):
    r = getattr(_norm, f'ta_{returns}')(df) if returns is not None else df

    def model(x):
        model = arch_model(x, p=1, q=1, dist='StudentsT', rescale=True)
        res = model.fit(update_freq=0, disp='off', show_warning=False)

        return res.forecast(horizon=forecast, method='analytic').variance.iloc[-1, -1] / (res.scale * res.scale)

    return r\
        .rolling(period)\
        .apply(model)\
        .rename(f"{df.name}_garch11({period})->{forecast}")


# rough stochastic fractional volatility
def ta_rsfv(df: Typing.PatchedDataFrame, period=255*2, lags=30, delta=1, open="Open", high="High", low="Low", close="Close") -> Typing.PatchedPandas:
    from pandas_ta_quant.technical_analysis import ta_gkyz_volatility, ta_vola_hurst
    rv = ta_gkyz_volatility(df, period=1, open=open, high=high, low=low, close=close)
    dfh = ta_vola_hurst(df, period=period, lags=lags)

    def fc(rv, h, nu, delta=1):
        i = np.arange(len(rv))
        cf = 1. / ((i + 1. / 2.) ** (h + 1. / 2.) * (i + 1. / 2. + delta))
        cf = np.fliplr([cf])[0]

        ldata = np.log(rv)
        fcst = (ldata * cf).sum() / sum(cf)

        def c_tilde(h):
            return scsp.gamma(3. / 2. - h) / scsp.gamma(h + 1. / 2.) * scsp.gamma(2. - 2. * h)

        return np.exp(fcst + 2 * nu ** 2 * c_tilde(h) * delta ** (2 * h))

    partials = dfh.apply(lambda r: partial(fc, h=r[0], nu=r[1], delta=delta), axis=1)
    return rv.rolling(lags).apply(lambda r: partials[r.index[-1]](r)).rename(f"RSFV_{period}_{lags}_{delta}")





from unittest import TestCase

import numpy as np

from pandas_ta_quant.technical_analysis import ta_garch11
from pandas_ta_quant_test.config import DF_TEST


class TestVolatilityForecast(TestCase):

    def test_garch11(self):
        df = DF_TEST[-220:].copy()

        garch11 = ta_garch11(df[["Close"]])

        print(garch11)
        np.testing.assert_almost_equal(garch11.iloc[-1, -1], 0.000040, 5)




from .bands import *
from .covariances import *
from .encoders import *
from .filters import *
from .edge_detect import *
from .forecast import *
from .indicators import *
from .normalizer import *
from .math import *


# create convenient type hint
import numpy as _np
import pandas as _pd

from pandas_ml_common import Typing
from pandas_ta_quant._decorators import for_each_top_level_row


def _normalize(df, convert_to):
    data = df.copy()

    if convert_to == 'returns':
        data = df.pct_change()
    if convert_to == 'log-returns':
        data = _np.log(df) - _np.log(df.shift(1))

    return data


@for_each_top_level_row
def ta_ewma_covariance(df: Typing.PatchedPandas, alpha=0.97, convert_to='returns', **kwargs):
    data = _normalize(df, convert_to)
    data.columns = data.columns.to_list()
    return data.ewm(com=alpha).cov()


@for_each_top_level_row
def ta_moving_covariance(df: Typing.PatchedPandas, period=30, convert_to='returns', **kwargs):
    data = _normalize(df, convert_to)
    data.columns = data.columns.to_list()
    return data.rolling(period).cov()


@for_each_top_level_row
def ta_mgarch_covariance(df: Typing.PatchedPandas, period=30, convert_to='returns', forecast=1, dist='t', **kwargs):
    assert period >= df.shape[1], f"period need to be > {df.shape[1]}"
    data = _normalize(df, convert_to)
    data.columns = data.columns.to_list()

    def mgarch_cov(window):
        import mgarch
        vol = mgarch.mgarch(dist)
        vol.fit(window.values)
        return _pd.DataFrame(
            vol.predict(forecast)["cov"],
            index=_pd.MultiIndex.from_product([[window.index[-1]], window.columns.to_list()]),
            columns=window.columns.to_list()
        )

    return _pd.concat([mgarch_cov(data.iloc[i - period: i]) for i in range(period, len(data) + 1)], axis=0)


@for_each_top_level_row
def ta_sparse_covariance(df: Typing.PatchedPandas, convert_to='returns', covariance='ewma', cov_arg=0.97, rho=0.1, inverse=False, **kwargs):
    from sklearn.covariance import graphical_lasso

    if covariance in ['ewma', 'weighted']:
        cov_func = ta_ewma_covariance
    elif covariance in ['rolling', 'moving']:
        cov_func = ta_moving_covariance
    elif covariance in ['garch', 'mgarch']:
        cov_func = ta_mgarch_covariance
    else:
        raise ValueError("unknown covariance expected one of [ewma, moving]")

    return \
        cov_func(df, cov_arg, convert_to) \
        .groupby(level=0) \
        .apply(lambda x: x if x.isnull().values.any() else \
                         _pd.DataFrame(graphical_lasso(x.values, rho, **kwargs)[int(inverse)], index=x.index, columns=x.columns))








from abc import abstractmethod
from collections import defaultdict
from datetime import timedelta
from typing import Union, Tuple, List

import pandas as pd
from bintrees import BinaryTree

from pandas_ml_common.utils.time_utils import parse_timestamp


class AbstractPriceTimeSeries():

    @abstractmethod
    def get_price(self, instrument: str, timestamp: Union[pd.Timestamp, str, Tuple[str, str]], currency: str = 'USD', convert_to: str = None):
        pass


class PriceTimeSeries(AbstractPriceTimeSeries):
    """
    entweder senden wir einen timestamp zusammen mit bid/ask, oder einer Quote mit spread (default 0).
    man kann auch einen bar (aka candle) senden mit open und close timestamps. default wird ein bar aber so übersetzt:

        open = 00:00:00.0001
        close = 00:00:00.0000 am nächsten Tag!

    wenn wir den Trade mit 23.59.59 machen dann ist der aktuellste Kurs der Open des Tages. was eigentlich logisch ist,
    bei einer ein datenquelle bekommen wir den Kurs erst am Ende des Tages. der früheste Trade kann also nur morgen mit
    dem open stattfinden. trades um Mitternacht werden mit dem close des Vortages ausgeführt.

    bei jeder Strategie muss man jetzt überlegen. wenn ich das Signal heute bekomme (eod). dann kann ich frühestens
    heute Morgen den Trade platzieren. entweder zum open Morgen also irgendwann nach Mitternacht. oder zum close von
    heute (angenommen man hat ein realtime feed) also morgen um exakt Mitternacht.

    idealerweise bauen wir ein library das Signale in trades übersetzt damit wir die für den user etwas einfacher
    abstrahieren können.
    """

    @staticmethod
    def from_dataframe(df: pd.DataFrame,
                       instrument: str,
                       ohlc: List[str] = None, bid_ask: List[str] = None, price: str = None,
                       currency: str = 'USD',
                       **kwargs):
        assert ohlc is not None or bid_ask is not None or price is not None, \
               "You need to pass one of ohlc, bid_ask oder price column name(s)"

        prices = PriceTimeSeries()

        def load_ohlc(row):
            prices.push_bar(instrument, row.name, *row[ohlc], currency=currency, **kwargs)

        def load_quote(row):
            prices.push_quote(instrument, row.name, *row[bid_ask], currency=currency)

        def load_price(row):
            prices.push_price(instrument, row.name, row[price], currency=currency, **kwargs)

        if ohlc is not None:
            func = load_ohlc
        elif bid_ask is not None:
            func = load_quote
        else:
            func = load_price

        df.apply(func, axis=1)
        return prices

    def __init__(self):
        self.timeseries = defaultdict(BinaryTree)

    def push_bar(self,
                 instrument: str,
                 date: Union[pd.Timestamp, str, Tuple[str, str]],
                 open: float,
                 high: float = None,
                 low: float = None,
                 close: float = None,
                 volume: float = None,
                 currency: str = 'USD',
                 open_time: str = '00:00:00.0001',
                 close_time: str = '00:00:00.0000',
                 spread: float = 0,
                 ):
        assert close != None, 'close price need to be passed'

        # high, low and volume are actually not used at all it is really just for convenience
        open_timestamp = parse_timestamp(f'{date} {open_time}')
        close_timestamp = parse_timestamp(f'{date} {close_time}')

        if close_timestamp < open_timestamp:
            close_timestamp = close_timestamp + timedelta(days=1)

        self.push_price(instrument, open_timestamp, open, currency, spread)
        self.push_price(instrument, close_timestamp, close, currency, spread)

    def push_price(self,
                   instrument: str,
                   timestamp: Union[pd.Timestamp, str, Tuple[str, str]],
                   price: float,
                   currency: str = 'USD',
                   spread: float = 0,
                  ):
        assert 0 <= spread <= 1, 'Spread need to be between 0 and 1'
        bid = price * (1 - spread / 2)
        ask = price * (1 + spread / 2)
        self.push_quote(instrument, timestamp, bid, ask, currency)

    def push_quote(self,
                   instrument: str,
                   timestamp: Union[pd.Timestamp, str, Tuple[str, str]],
                   bid: float,
                   ask: float,
                   currency: str = 'USD',
                  ):
        tst = parse_timestamp(timestamp)
        self.timeseries[(instrument, currency)][tst] = bid, ask

    def get_price(self, instrument: str, timestamp: Union[pd.Timestamp, str, Tuple[str, str]], currency: str = 'USD', convert_to: str = None):
        tst = parse_timestamp(timestamp)
        price = self.timeseries[(instrument, currency)].floor_item(tst)

        if convert_to is not None and convert_to != currency:
            fxrate = 1 # TODO later we want to do a recurse lookup for an fx rate: self.get_price(convert_to, price[0], currency)
            price = (price[0], (price[1][0] * fxrate, price[1][1] * fxrate))
            raise NotImplementedError("Multiple currencies not supported at the moment")

        return price



"""Augment pandas DataFrame with methods for machine learning"""
__version__ = '0.2.4'
import os
import re

from setuptools import setup, find_packages
from setuptools.command.sdist import sdist

URL = 'https://github.com/KIC/pandas-ml-quant'
NAME = 'pandas-ta-quant'


class SDist(sdist):

    def fix_github_links(self, lines):
        #   ist  https://github.com/KIC/pandas-ml-quant/pandas-ml-common/tree/0.2.0/./examples/
        #   soll https://github.com/KIC/pandas-ml-quant/tree/0.2.0/pandas-ml-common/./examples/
        def fix_line(line):
            fixed_images = re.sub(r'(^\[ghi\d+]:\s+)', f'\\1{URL}/raw/{__version__}/{NAME}/', line)
            fixed_location = re.sub(r'(^\[ghl\d+]:\s+)', f'\\1{URL}/tree/{__version__}/{NAME}/', fixed_images)
            fixed_files = re.sub(r'(^\[ghf\d+]:\s+)', f'\\1{URL}/blob/{__version__}/{NAME}/', fixed_location)
            return fixed_files

        return [fix_line(line) for line in lines]

    def make_release_tree(self, base_dir, files):
        # create the regular distribution files
        super().make_release_tree(base_dir, files)

        # but then fix the github links
        readme_file = os.path.join(base_dir, 'Readme.md')
        readme_lines = open(readme_file).readlines()

        with open(readme_file, 'w') as f:
            f.writelines(self.fix_github_links(readme_lines))


setup(
    name=NAME,
    version=__version__,
    author='KIC',
    author_email='',
    packages=find_packages(),
    scripts=[],
    url=f'{URL}/{NAME}',
    license='MIT',
    description=__doc__,
    long_description=open('Readme.md').read(),
    long_description_content_type='text/markdown',
    install_requires=open("requirements.frozen.txt").read().splitlines() + [f"pandas-ml-common=={__version__}"],
    extras_require={
       "dev": open("dev-requirements.frozen.txt").read().splitlines(),
    },
    include_package_data=True,
    classifiers=[
        # Chose either "3 - Alpha", "4 - Beta" or "5 - Production/Stable" as the current state of your package
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Build Tools',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.8',
    ],
    keywords=['pandas', 'ml', 'util', 'quant'],
    cmdclass={
        'sdist': SDist
    },
)



from unittest import TestCase

import pandas as pd

from pandas_ta_quant import _TA
from pandas_ta_quant_test.config import DF_TEST_MULTI, DF_TEST_MULTI_ROW_MULTI_COLUMN


class TestTheAllIndicator(TestCase):

    def test_all_indicator_multi_columns(self):
        mdf = DF_TEST_MULTI.copy()

        # test does not throw any exception
        # TechnicalAnalysis(df).all()
        res = _TA(mdf).all()

        print(res.tail())
        print(res.columns)
        self.assertEqual(3, res.columns.nlevels)
        self.assertEqual(3835, len(res))

    def test_all_indicator_multirows(self):
        mdf = DF_TEST_MULTI_ROW_MULTI_COLUMN.copy()

        # test does not throw any exception
        # TechnicalAnalysis(df).all()
        res = _TA(mdf).all()

        print(res.tail())
        print(res.columns)
        self.assertEqual(3, res.columns.nlevels)
        self.assertEqual(2, res.index.nlevels)
        self.assertEqual(3835 * 2, len(res))

        pd.testing.assert_frame_equal(res.loc["A"], res.loc["B"])




__version__ = '0.2.4'

import os
import shutil
from pathlib import Path

import nox

nox.options.envdir = '../.nox'


@nox.session(python=["3.8"], reuse_venv=False, venv_params=['--system-site-packages'])
def tests(session):
    dist_file = f"/tmp/pandas-ta-quant-{__version__}.zip"

    # install testing requirements and local dependencies
    session.install("-r", "dev-requirements.txt")
    session.install("-r", "requirements.txt")
    session.install(f"/tmp/pandas-ml-common-{__version__}.zip")

    # create distribution and install
    session.run("python", "setup.py", "sdist",  "-d", "/tmp/", "--formats=zip", env={})
    session.install(dist_file, "--no-deps")

    # create notebook kernels
    kernel = "ta_quant"
    session.run("python", "-m", "ipykernel", "install", "--user", "--name", kernel, "--display-name", f"{kernel} py38")

    # run tests
    session.run("python", "-m", "unittest", "discover", env={"TOX_KERNEL": kernel})

    # clean up kernels
    kernels_home_dir = os.path.join(Path.home(), ".local", "share", "jupyter", "kernels", kernel)
    if os.path.exists(kernels_home_dir):
        print(f"clean up dir {kernels_home_dir}")

        try:
            shutil.rmtree(kernels_home_dir)
        except Exception as e:
            print(e)

    # clean up egg info
    shutil.rmtree("pandas_ta_quant.egg-info")

    # make link check
    session.run("python", "pandas_ta_quant_test/check_links.py", dist_file)

    # freeze versions and rebuild with frozen versions
    session.run("python", "../freeze_versions.py", "requirements.txt", "dev-requirements.txt")
    session.run("python", "setup.py", "sdist",  "-d", "/tmp/", "--formats=zip", env={})


from functools import wraps
from typing import List

import pandas as pd
import logging
import pandas_ta_quant.technical_analysis as technichal_analysis

_log = logging.getLogger(__name__)


class TechnicalAnalysis(object):
    _ta_indicator_info = {}

    def __init__(self, df: pd.DataFrame):
        self.df = df

    @property
    def help(self):
        return pd.DataFrame(TechnicalAnalysis._ta_indicator_info).T.rename(columns={0: 'module', 1: "doc"})


# add wrapper to call all indicators on data frames
def wrapper(func):
    @wraps(func)
    def wrapped(quant, *args, **kwargs):
        return func(quant.df, *args, **kwargs)

    return wrapped


# add indicators
def patch_indicators(patch_type, indicators: List["__init__.py"]):
    for indicator_functions in indicators:
        for indicator_function in dir(indicator_functions):
            if indicator_function.startswith("ta_"):
                func = getattr(indicator_functions, indicator_function)
                setattr(patch_type, indicator_function[3:], wrapper(func))
                if hasattr(patch_type, "_ta_indicator_info"):
                    patch_type._ta_indicator_info[indicator_function] = [func.__module__, func.__doc__]

patch_indicators(TechnicalAnalysis, [technichal_analysis])


# add plotting
try:
    from pandas_ta_quant_plot import PlotContext
    setattr(TechnicalAnalysis, "plot", wrapper(PlotContext))
except Exception as e:
    _log.warning(f"pandas_ta_quant_plot not found: {e}")


import warnings
from functools import partial
from typing import Iterable

import numpy as np
import logging
from pandas_ta_quant._decorators import *
from pandas_ml_common import Typing
from statsmodels.tsa.statespace.sarimax import SARIMAX
from statsmodels.tools.sm_exceptions import ConvergenceWarning
warnings.simplefilter('ignore', ConvergenceWarning)

_log = logging.getLogger(__name__)


@for_each_top_level_row
@for_each_column
def ta_hmm(df: Typing.PatchedSeries, nr_components, means_estimator, period=250, forecast_period=1, n_iter=500):
    from hmmlearn import hmm

    X = df.to_frame().pct_change(forecast_period).dropna().values

    def moving_hmm(X):
        # new model
        model = hmm.GaussianHMM(n_components=nr_components, n_iter=n_iter, covariance_type="diag", init_params="tc")

        # set means
        model.means_ = np.array(means_estimator(X)).reshape((nr_components, X.shape[1]))

        # set initial state probability
        model.startprob_ = np.zeros(nr_components)
        model.startprob_[np.abs(model.means_ - X[0]).argmin()] = 1.0

        # fit model
        model = model.fit(X)

        # predict current
        hidden_state_probabilities = model.predict_proba(X)
        probabilities = hidden_state_probabilities[-1:]
        transition_matrix = model.transmat_
        next_state_probabilities = probabilities @ transition_matrix

        return next_state_probabilities.reshape(nr_components)

    p = period - 1
    estimates = np.array([moving_hmm(X[i - p:i]) for i in range(p, len(X))])
    print(estimates.shape)
    return pd.DataFrame(estimates, index=df.index[p+forecast_period:])


@for_each_top_level_row
@for_each_column
def ta_sarimax(df: Typing.PatchedSeries, period=60, forecast=1, order=(1, 0, 1), alpha=0.5):
    assert forecast > 0, "forecast need to be > 0"
    forecast = forecast if isinstance(forecast, Iterable) else [forecast]
    res = pd.DataFrame({}, index=df.index)
    dfclean = df.dropna()

    def arima(x, fc):
        try:
            forecasted = SARIMAX(x, order=order).fit(disp=-1).forecast(fc, alpha=alpha)
        except Exception as e:
            _log.warning(f"failed arma model: {e}")
            forecasted = np.nan

        return forecasted  # TODO eventually return conficence as well, conf_int[:, 0], conf_int[:, 1]

    for fc in forecast:
        res = res.join(dfclean.rolling(period).apply(partial(arima, fc=fc), raw=True).rename(f'{df.name}_sarimax_fc_{fc}'))

    return res






from typing import Union as _Union

# create convenient type hint
import numpy as _np
import pandas as _pd

from pandas_ml_common import Typing as _t
from pandas_ml_common.utils import inner_join, add_multi_index
from pandas_ta_quant._decorators import *
from pandas_ta_quant._utils import wilders_smoothing as _ws, with_column_suffix as _wcs

_PANDAS = _Union[_pd.DataFrame, _pd.Series]


@for_each_top_level_row
def ta_sma(df: _PANDAS, period=12) -> _PANDAS:
    return _wcs(f"sma_{period}", df.rolling(period).mean())


@for_each_top_level_row
def ta_ema(df: _PANDAS, period=12) -> _PANDAS:
    return _wcs(f"ema_{period}", df.ewm(span=period, adjust=False, min_periods=period-1).mean())


@for_each_top_level_row
def ta_wilders(df: _PANDAS, period=12) -> _PANDAS:
    # NOTE this does not work with @for_each_column decorator!
    if df.ndim > 1:
        resdf = _pd.DataFrame({}, index=df.index)
        for col in df.columns:
            s = df[col].dropna()
            res = _np.zeros(s.shape)
            _ws(s.values, period, res)
            resdf = resdf.join(_pd.DataFrame({col: res}, index=s.index))

        res = resdf
    else:
        res = ta_wilders(df.to_frame(), period).iloc[:, 0]

    return _wcs(f"wilders_{period}", res)


@for_each_top_level_row
def ta_multi_bbands(s: _pd.Series, period=12, stddevs=[0.5, 1.0, 1.5, 2.0], ddof=1, include_mean=True) -> _PANDAS:
    assert s.ndim == 1, "Expected Series not Frame"
    mean = s.rolling(period).mean().rename("mean")
    std = s.rolling(period).std(ddof=ddof)
    df = _pd.DataFrame({}, index=mean.index)

    for stddev in reversed(stddevs):
        df[f'lower-{stddev}'] = mean - (std * stddev)

    if include_mean:
        df["mean"] = mean

    for stddev in stddevs:
        df[f'upper-{stddev}'] = mean + (std * stddev)

    return df


@for_each_top_level_row
def ta_multi_ma(df: _t.PatchedDataFrame, average_function='sma', period=12, factors=_np.linspace(1 - 0.2, 1 + 0.2, 5)) -> _t.PatchedDataFrame:
    ma = {'sma': ta_sma, 'ema': ta_ema, 'wilder': ta_wilders}
    res = _pd.DataFrame({}, index=df.index)

    if df.ndim > 1:
        res = None
        for col in df.columns.to_list():
            _df = ta_multi_ma(df[col], average_function, period, factors)
            _df = add_multi_index(_df, col)
            res = inner_join(res, _df, force_multi_index=True)

        return res

    for x in factors:
        res[f"{df.name}_{average_function}({period})({x:.3f})"] = ma[average_function](df, period) * x

    return res





from unittest import TestCase

import talib

from pandas_ml_common import Constant
from pandas_ta_quant.technical_analysis import *
from pandas_ta_quant_test.config import DF_TEST


class TestIndicator(TestCase):

    def test__mean_returns(self):
        me = ta_mean_returns(DF_TEST[["Close", "Volume"]], 20)[-100:]
        np.testing.assert_array_almost_equal(me.iloc[-1].values, np.array([0.000810, 0.033791]), 5)

    def test__ema(self):
        me = ta_ema(DF_TEST["Close"], 20)[-100:]
        ta = talib.EMA(DF_TEST["Close"], 20)[-100:]

        np.testing.assert_array_almost_equal(me, ta)

    def test__wilders(self):
        me = ta_wilders(DF_TEST["Close"], 20)[-100:]
        ta = talib.EMA(DF_TEST["Close"], 20 * 2 -1)[-100:]

        np.testing.assert_array_almost_equal(me, ta)

    def test__macd(self):
        my_macd = ta_macd(DF_TEST["Close"], relative=False)
        talib_macd = pd.DataFrame(talib.MACD(DF_TEST["Close"])).T

        np.testing.assert_array_almost_equal(talib_macd[-100:].values[0], my_macd[-100:].values[0])

        # test multi
        my_multi_macd = ta_macd(DF_TEST[["Close", "Open"]], relative=False)
        self.assertListEqual([('Close', 'macd_12,26,9'),
                              ('Close', 'signal_12,26,9'),
                              ('Close', 'histogram_12,26,9'),
                              ('Open', 'macd_12,26,9'),
                              ('Open', 'signal_12,26,9'),
                              ('Open', 'histogram_12,26,9')],
                             my_multi_macd.columns.to_list())

    def test__mom(self):
        me = ta_mom(DF_TEST["Close"], period=10, relative=False)[-100:]
        ta = talib.MOM(DF_TEST["Close"])[-100:]

        np.testing.assert_array_almost_equal(me, ta)

    def test__stddev(self):
        me = ta_stddev(DF_TEST["Close"], period=5, ddof=0, downscale=False)[-100:]
        ta = talib.STDDEV(DF_TEST["Close"])[-100:]

        np.testing.assert_array_almost_equal(me, ta)

    def test__apo(self):
        me = ta_apo(DF_TEST["Close"], relative=False)[-100:]
        ta = talib.APO(DF_TEST["Close"])[-100:]

        np.testing.assert_array_almost_equal(me, ta)

    def test__trix(self):
        me = ta_trix(DF_TEST["Close"])[-100:]
        ta = talib.TRIX(DF_TEST["Close"])[-100:]

        np.testing.assert_array_almost_equal(me, ta)

    def test__tr(self):
        me = ta_tr(DF_TEST, relative=False)[-100:]
        ta = talib.TRANGE(DF_TEST["High"], DF_TEST["Low"], DF_TEST["Close"])[-100:]

        np.testing.assert_array_almost_equal(me, ta)

    def test__atr(self):
        me = ta_atr(DF_TEST, relative=False)[-100:]
        ta = talib.ATR(DF_TEST["High"], DF_TEST["Low"], DF_TEST["Close"])[-100:]

        np.testing.assert_array_almost_equal(me, ta)

    def test__adx(self):
        me = ta_adx(DF_TEST, relative=False)[-100:]
        ta_pdi = talib.PLUS_DI(DF_TEST["High"], DF_TEST["Low"], DF_TEST["Close"])[-100:]
        ta_mdi = talib.MINUS_DI(DF_TEST["High"], DF_TEST["Low"], DF_TEST["Close"])[-100:]

        ta_pdm = talib.PLUS_DM(DF_TEST["High"], DF_TEST["Low"])[-100:]
        ta_mdm = talib.MINUS_DM(DF_TEST["High"], DF_TEST["Low"])[-100:]
        ta_dx = talib.ADX(DF_TEST["High"], DF_TEST["Low"], DF_TEST["Close"])[-100:]

        np.testing.assert_array_almost_equal(me["ADX"], ta_dx / 100)
        np.testing.assert_array_almost_equal(me["+DI"], ta_pdi / 100)
        np.testing.assert_array_almost_equal(me["-DI"], ta_mdi / 100)

        # I have absolutely no idea why we are off by a constant factor. but since it is a constant factor we can
        # neglect it
        np.testing.assert_array_almost_equal(me["+DM"], ta_pdm * 0.07142861354566638)
        np.testing.assert_array_almost_equal(me["-DM"], ta_mdm * 0.07142861354566638)

    def test__bbands(self):
        me = ta_bbands(DF_TEST["Close"], period=5, ddof=0)[-100:]
        u, m, l = talib.BBANDS(DF_TEST["Close"])
        u = u[-100:]
        m = m[-100:]
        l = l[-100:]

        np.testing.assert_array_almost_equal(me["mean"], m)
        np.testing.assert_array_almost_equal(me["upper"], u)
        np.testing.assert_array_almost_equal(me["lower"], l)

    def test__rsi(self):
        me = ta_rsi(DF_TEST["Close"], period=14)[-100:]
        ta = talib.RSI(DF_TEST["Close"])[-100:]

        np.testing.assert_array_almost_equal(me, ta / 100)

    def test__williams_R(self):
        me = ta_williams_R(DF_TEST)[-100:]
        ta = talib.WILLR(DF_TEST["High"], DF_TEST["Low"], DF_TEST["Close"])[-100:]

        np.testing.assert_array_almost_equal(me, ta / -100)

    def test__ultimate(self):
        me = ta_ultimate_osc(DF_TEST)[-100:]
        ta = talib.ULTOSC(DF_TEST["High"], DF_TEST["Low"], DF_TEST["Close"])[-100:]

        np.testing.assert_array_almost_equal(me, ta / 100)

    def test__ppo(self):
        me = ta_ppo(DF_TEST["Close"], exponential=False)[-100:]
        ta = talib.PPO(DF_TEST["Close"])[-100:]

        np.testing.assert_array_almost_equal(me, ta / 100)

    def test__bop(self):
        me = ta_bop(DF_TEST)[-100:]
        ta = talib.BOP(DF_TEST["Open"], DF_TEST["High"], DF_TEST["Low"], DF_TEST["Close"])[-100:]

        np.testing.assert_array_almost_equal(me, ta)

    def test__cci(self):
        me = ta_cci(DF_TEST)[-100:]
        ta = talib.CCI(DF_TEST["High"], DF_TEST["Low"], DF_TEST["Close"])[-100:]

        np.testing.assert_array_almost_equal(me, ta / 100)

    def test__up_down_volatility_ratio(self):
        me = ta_up_down_volatility_ratio(DF_TEST)[-1:]

        np.testing.assert_array_almost_equal(me, np.array([[-0.067, -0.232, -0.073, -0.318, -0.319,  1.776]]), 0.001)

    def test__z_score(self):
        me = ta_zscore(DF_TEST, downscale=False)[-1:]

        np.testing.assert_array_almost_equal(me, np.array([[ 0.546, 0.256, 0.271, 0.457,  0.457, -0.929]]), 0.001)

    def test__ta_multi_bbands(self):
        me = ta_multi_bbands(DF_TEST["Close"])[-1:]
        print(me.values.round(3))

        np.testing.assert_array_almost_equal(
            me,
            np.array([[308.406, 309.257, 310.108, 310.959, 311.81, 312.661, 313.512, 314.363, 315.214]]),
            0.001)

    def test_ta_sinusoidial_weekday(self):
        me = ta_sinusoidal_week_day(DF_TEST["Close"])[-1:]

        np.testing.assert_almost_equal(me.values[0], 1.2246467991473532e-16, 6)

    def test_ta_sinusoidial_week(self):
        me = ta_sinusoidal_week(DF_TEST["Close"])

        np.testing.assert_almost_equal(me.values[-1], -0.35460488704253595, 4)
        np.testing.assert_almost_equal(me.loc["2017-01-03"], 0.12053668025532306, 4)
        np.testing.assert_almost_equal(me.loc["2017-12-29"], -2.4492935982947064e-16, 4)

    def test_poly_coeff(self):
        me = ta_poly_coeff(DF_TEST["Close"], period=16)
        print(me.tail())

        np.testing.assert_almost_equal(
            me.iloc[-1].values,
            np.array([-3.279875, 4.177750, 0.183198]),
            6
        )

    def test_slope(self):
        me = ta_slope(DF_TEST[["Close", "Open"]], period=20)
        # print(me.tail())

        np.testing.assert_almost_equal(
            [0.225188, 0.267564],
            me.iloc[-1].values,
            6
        )

    def test_hurst(self):
        h = ta_vola_hurst(DF_TEST[-255*2:])

        np.testing.assert_array_almost_equal(
            [0.147095, 0.473749],
            h.iloc[-1].values
        )

    def test_potential_turning_point(self):
        tps3 = ta_potential_turning_point(DF_TEST["Close"], 120, edge_detector='poly', edge_period=3)
        tps1 = ta_potential_turning_point(DF_TEST["Close"], 120, edge_detector='naive', edge_period=3)
        tps2 = ta_potential_turning_point(DF_TEST["Close"], 120, edge_detector='mean', edge_period=3)

        print(tps1.sum())
        print(tps2.sum())
        print(tps3.sum())
